"""
Parameter conversion table

*CONVERSION_TABLE* gives the old model name and a dictionary of old parameter
names for each parameter in sasmodels.  This is used by :mod:`convert` to
determine the equivalent parameter set when comparing a sasmodels model to
the models defined in previous versions of SasView and sasmodels. This is now
versioned based on the version number of SasView.

When any sasmodels parameter or model name is changed, this must be modified to
account for that.

Usage::

    <old_Sasview_version> : {
        <new_model_name> : [
            <old_model_name> ,
            {
                <new_param_name_1> : <old_param_name_1>,
                ...
                <new_param_name_n> : <old_param_name_n>
            }
        ]
    }

Any future parameter and model name changes can and should be given in this
table for future compatibility.
"""

CONVERSION_TABLE = {
    (3,1,2) : {
    "adsorbed_layer": [
        "Core2ndMomentModel",
        {
            "scale": "scale",
            "second_moment": "second_moment",
            "density_shell": "density_poly",
            "sld_solvent": "sld_solv",
            "radius": "radius_core",
            "volfraction": "volf_cores",
            "background": "background",
            "adsorbed_amount": "ads_amount",
            "sld_shell": "sld_poly"
        }
    ],
    "barbell": [
        "BarBellModel",
        {
            "sld": "sld_barbell",
            "length": "len_bar",
            "radius_bell": "rad_bell",
            "radius": "rad_bar",
            "sld_solvent": "sld_solv"
        }
    ],
    "bcc_paracrystal": [
        "BCCrystalModel",
        {
            "sld": "sldSph",
            "sld_solvent": "sldSolv"
        }
    ],
    "be_polyelectrolyte": [
        "BEPolyelectrolyte",
        {
            "ionization_degree": "alpha",
            "polymer_concentration": "c",
            "salt_concentration": "cs",
            "virial_param": "h",
            "background": "background",
            "contrast_factor": "k",
            "bjerrum_length": "lb",
            "monomer_length": "b"
        }
    ],
    "binary_hard_sphere": [
        "BinaryHSModel",
        {
            "sld_sm": "ss_sld",
            "sld_lg": "ls_sld",
            "volfraction_sm": "vol_frac_ss",
            "radius_lg": "l_radius",
            "radius_sm": "s_radius",
            "volfraction_lg": "vol_frac_ls",
            "sld_solvent": "solvent_sld"
        }
    ],
    "broad_peak": [
        "BroadPeakModel",
        {
            "peak_pos": "q_peak",
            "scale": None,
            "lorentz_length": "length_l",
            "porod_scale": "scale_p",
            "lorentz_exp": "exponent_l",
            "lorentz_scale": "scale_l",
            "porod_exp": "exponent_p"
        }
    ],
    "capped_cylinder": [
        "CappedCylinderModel",
        {
            "sld": "sld_capcyl",
            "length": "len_cyl",
            "radius_cap": "rad_cap",
            "radius": "rad_cyl",
            "sld_solvent": "sld_solv"
        }
    ],
    "core_multi_shell": [
        "CoreMultiShellModel",
        {
            "thickness": "thick_shell",
            "sld": "sld_shell",
            "radius": "rad_core0",
            "sld_core": "sld_core0",
            "sld_solvent": "sld_solv",
            "n": "n_shells",
            "M0:sld_core": "M0_sld_core0",
            "mtheta:sld_core": "M_theta_core0",
            "mphi:sld_core": "M_phi_core0",
            "M0:sld1": "M0_sld_shell1",
            "mtheta:sld1": "M_theta_shell1",
            "mphi:sld1": "M_phi_shell1",
            "M0:sld2": "M0_sld_shell2",
            "mtheta:sld2": "M_theta_shell2",
            "mphi:sld2": "M_phi_shell2",
            "M0:sld3": "M0_sld_shell3",
            "mtheta:sld3": "M_theta_shell3",
            "mphi:sld3": "M_phi_shell3",
            "M0:sld4": "M0_sld_shell4",
            "mtheta:sld4": "M_theta_shell4",
            "mphi:sld4": "M_phi_shell4",
            "M0:sld_solvent": "M0_sld_solv",
            "mtheta:sld_solvent": "M_theta_solv",
            "mphi:sld_solvent": "M_phi_solv",
            "up:frac_i": "Up_frac_i",
            "up:frac_f": "Up_frac_f",
            "up:angle": "Up_theta",
        }
    ],
    "core_shell_bicelle": [
        "CoreShellBicelleModel",
        {
            "phi": "axis_phi",
            "sld_core": "core_sld",
            "sld_rim": "rim_sld",
            "thick_face": "face_thick",
            "sld_solvent": "solvent_sld",
            "thick_rim": "rim_thick",
            "sld_face": "face_sld",
            "theta": "axis_theta"
        }
    ],
    "core_shell_cylinder": [
        "CoreShellCylinderModel",
        {
            "theta": "axis_theta",
            "phi": "axis_phi",
            "sld_shell": "shell_sld",
            "sld_solvent": "solvent_sld",
            "sld_core": "core_sld"
        }
    ],
    "core_shell_ellipsoid:1": [
        "CoreShellEllipsoidModel",
        {
            "sld_core": "sld_core",
            "sld_shell": "sld_shell",
            "sld_solvent": "sld_solvent",
            "radius_equat_core": "equat_core",
            "x_core": "polar_core",
            "thick_shell": "equat_shell",
            "x_polar_shell": "polar_shell",
            "theta": "axis_theta",
            "phi": "axis_phi",
        }
    ],
    "core_shell_ellipsoid": [
        "CoreShellEllipsoidXTModel",
        {
            "sld_core": "sld_core",
            "sld_shell": "sld_shell",
            "sld_solvent": "sld_solvent",
            "radius_equat_core": "equat_core",
            "thick_shell": "T_shell",
            "x_core": "X_core",
            "x_polar_shell": "XpolarShell",
            "theta": "axis_theta",
            "phi": "axis_phi",
        }
    ],
    "core_shell_parallelepiped": [
        "CSParallelepipedModel",
        {
            "sld_core": "sld_pcore",
            "sld_a": "sld_rimA",
            "sld_b": "sld_rimB",
            "sld_c": "sld_rimC",
            "sld_solvent": "sld_solv",
            "length_a": "shortA",
            "length_b": "midB",
            "length_c": "longC",
            "thick_rim_a": "rimA",
            "thick_rim_c": "rimC",
            "thick_rim_b": "rimB",
            "theta": "parallel_theta",
            "phi": "parallel_phi",
            "psi": "parallel_psi",
        }
    ],
    "core_shell_sphere": [
        "CoreShellModel",
        {
            "sld_core": "core_sld",
            "sld_shell": "shell_sld",
            "sld_solvent": "solvent_sld",
            "M0:sld_core": "M0_sld_core",
            "mtheta:sld_core": "M_theta_core",
            "mphi:sld_core": "M_phi_core",
            "M0:sld_shell": "M0_sld_shell",
            "mtheta:sld_shell": "M_theta_shell",
            "mphi:sld_shell": "M_phi_shell",
            "M0:sld_solvent": "M0_sld_solv",
            "mtheta:sld_solvent": "M_theta_solv",
            "mphi:sld_solvent": "M_phi_solv",
            "up:frac_i": "Up_frac_i",
            "up:frac_f": "Up_frac_f",
            "up:angle": "Up_theta"
        }
    ],
    "correlation_length": [
        "CorrLength",
        {
            "porod_scale": "scale_p",
            "lorentz_scale": "scale_l",
            "porod_exp": "exponent_p",
            "lorentz_exp": "exponent_l",
            "cor_length": "length_l"
        },
        "CorrLengthModel"
    ],
    "cylinder": [
        "CylinderModel",
        {
            "sld": "sldCyl",
            "theta": "cyl_theta",
            "phi": "cyl_phi",
            "sld_solvent": "sldSolv",
            "M0:sld": "M0_sld_cyl",
            "mtheta:sld": "M_theta_cyl",
            "mphi:sld": "M_phi_cyl",
            "M0:sld_solvent": "M0_sld_solv",
            "mtheta:sld_solvent": "M_theta_solv",
            "mphi:sld_solvent": "M_phi_solv",
            "up:frac_i": "Up_frac_i",
            "up:frac_f": "Up_frac_f",
            "up:angle": "Up_theta"
        }
    ],
    "dab": [
        "DABModel",
        {
            "cor_length": "length"
        }
    ],
    "ellipsoid": [
        "EllipsoidModel",
        {
            "phi": "axis_phi",
            "radius_equatorial": "radius_b",
            "sld": "sldEll",
            "theta": "axis_theta",
            "radius_polar": "radius_a",
            "sld_solvent": "sldSolv"
        }
    ],
    "elliptical_cylinder": [
        "EllipticalCylinderModel",
        {
            "axis_ratio": "r_ratio",
            "radius_minor": "r_minor",
            "sld": "sldCyl",
            "sld_solvent": "sldSolv",
            "theta": "cyl_theta",
            "phi": "cyl_phi",
            "psi": "cyl_psi",
        }
    ],
    "fcc_paracrystal": [
        "FCCrystalModel",
        {
            "sld": "sldSph",
            "sld_solvent": "sldSolv"
        }
    ],
    "flexible_cylinder": [
        "FlexibleCylinderModel",
        {
            "sld": "sldCyl",
            "sld_solvent": "sldSolv"
        }
    ],
    "flexible_cylinder_elliptical": [
        "FlexCylEllipXModel",
        {
            "sld": "sldCyl",
            "sld_solvent": "sldSolv"
        }
    ],
    "fractal": [
        "FractalModel",
        {
            "sld_block": "sldBlock",
            "radius": "radius",
            "cor_length": "cor_length",
            "sld_solvent": "sldSolv",
            "fractal_dim": "fractal_dim"
        }
    ],
    "fractal_core_shell": [
        "FractalCoreShell",
        {
            "sld_core": "core_sld",
            "sld_shell": "shell_sld",
            "sld_solvent": "solvent_sld",
            "radius": "radius",
            "thickness": "thickness",
            "fractal_dim": "frac_dim",
            "cor_length": "cor_length",
            "volfraction": "volfraction",
        },
        "FractalCoreShellModel"
    ],
    "fuzzy_sphere": [
        "FuzzySphereModel",
        {
            "sld": "sldSph",
            "fuzziness": "fuzziness",
            "radius": "radius",
            "sld_solvent": "sldSolv"
        }
    ],
    "gauss_lorentz_gel": [
        "GaussLorentzGel",
        {
            "gauss_scale": "scale_g",
            "cor_length_dynamic": "dyn_colength",
            "cor_length_static": "stat_colength",
            "background": "background",
            "lorentz_scale": "scale_l"
        },
        "GaussLorentzGelModel"
    ],
    "gaussian_peak": [
        "Peak Gauss Model",
        {
            "peak_pos": "q0",
            "sigma": "B",
        },
        "PeakGaussModel",
    ],
    "gel_fit": [
        "GelFitModel",
        {
            "rg": "radius",
            "lorentz_scale": "lScale",
            "guinier_scale": "gScale",
            "fractal_dim": "FractalExp",
            "cor_length": "zeta",
        }
    ],
    "guinier": [
        "Guinier",
        {
            "rg": "rg"
        },
        "GuinierModel",
    ],
    "guinier_porod": [
        "GuinierPorod",
        {
            "s": "dim",
            "rg": "rg",
            "porod_exp": "m",
            "scale": "scale",
            "background": "background"
        },
        "GuinierPorodModel",
    ],
    "hardsphere": [
        "HardsphereStructure",
        {
            "scale": "scale_factor",
            "radius_effective": "effect_radius",
        }
    ],
    "hayter_msa": [
        "HayterMSAStructure",
        {
            "scale": "scale_factor",
            "radius_effective": "effect_radius",
            "volfraction": "volfraction",
            "charge": "charge",
            "temperature": "temperature",
            "concentration_salt": "saltconc",
            "dielectconst": "dielectconst",
        }
    ],
    "hollow_cylinder": [
        "HollowCylinderModel",
        {
            "sld": "sldCyl",
            "sld_solvent": "sldSolv",
            "radius": "core_radius",
            "thickness": "radius",
            "length": "length",
            "theta": "axis_theta",
            "phi": "axis_phi",
        }
    ],
    "hollow_rectangular_prism": [
        "RectangularHollowPrismModel",
        {
            "sld": "sldPipe",
            "sld_solvent": "sldSolv",
            "length_a": "short_side",
            "b2a_ratio": "b2a_ratio",
            "c2a_ratio": "c2a_ratio",
            "thickness": "thickness",
        }
    ],
    "hollow_rectangular_prism_thin_walls": [
        "RectangularHollowPrismInfThinWallsModel",
        {
            "sld": "sldPipe",
            "sld_solvent": "sldSolv",
            "length_a": "short_side",
            "b2a_ratio": "b2a_ratio",
            "c2a_ratio": "c2a_ratio",
        }
    ],
    "lamellar": [
        "LamellarModel",
        {
            "sld": "sld_bi",
            "sld_solvent": "sld_sol",
            "thickness": "bi_thick"
        }
    ],
    "lamellar_hg": [
        "LamellarFFHGModel",
        {
            "sld": "sld_tail",
            "sld_solvent": "sld_solvent",
            "sld_head": "sld_head",
            "length_tail": "t_length",
            "length_head": "h_thickness"
        }
    ],
    "lamellar_hg_stack_caille": [
        "LamellarPSHGModel",
        {
            "sld": "sld_tail",
            "sld_head": "sld_head",
            "sld_solvent": "sld_solvent",
            "length_tail": "deltaT",
            "length_head": "deltaH",
            "d_spacing": "spacing",
            "Caille_parameter": "caille",
            "Nlayers": "n_plates",
        }
    ],
    "lamellar_stack_caille": [
        "LamellarPSModel",
        {
            "sld": "sld_bi",
            "sld_solvent": "sld_sol",
            "thickness": "delta",
            "d_spacing": "spacing",
            "Caille_parameter": "caille",
            "Nlayers": "n_plates",
        }
    ],
    "lamellar_stack_paracrystal": [
        "LamellarPCrystalModel",
        {
            "sld": "sld_layer",
            "sld_solvent": "sld_solvent",
            "thickness": "thickness",
            "d_spacing": "spacing",
            "sigma_d": "pd_spacing",
            "Nlayers": "Nlayers",
        }
    ],
    "line": [
        "LineModel",
        {
            "slope": "B",
            "scale": None,
            "background": None,
            "intercept": "A"
        }
    ],
    "linear_pearls": [
        "LinearPearlsModel",
        {
            "sld": "sld_pearl",
            "sld_solvent": "sld_solv",
            "edge_sep": "edge_separation"
        }
    ],
    "lorentz": [
        "Lorentz",
        {
            "cor_length": "length"
        },
        "LorentzModel",
    ],
    "mass_fractal": [
        "MassFractalModel",
        {
            "cutoff_length": "co_length",
            "radius": "radius",
            "fractal_dim_mass": "mass_dim"
        }
    ],
    "mass_surface_fractal": [
        "MassSurfaceFractal",
        {
            "rg_cluster": "cluster_rg",
            "fractal_dim_mass": "mass_dim",
            "radius": "radius",
            "fractal_dim_surf": "surface_dim",
            "rg_primary": "primary_rg"
        }
    ],
    "mono_gauss_coil": [
        "Debye",
        {
            "rg": "rg",
            "i_zero": "scale",
            "background": "background",
        },
        "DebyeModel",
    ],
    "multilayer_vesicle": [
        "MultiShellModel",
        {
            "radius": "core_radius",
            "sld_solvent": "core_sld",
            "n_shells": "n_pairs",
            "thick_shell": "s_thickness",
            "sld": "shell_sld",
            "thick_solvent": "w_thickness",
        }
    ],
    "onion": [
        "OnionExpShellModel",
        {
            "n_shells": "n_shells",
            "A": "A_shell",
            "sld_core": "sld_core0",
            "radius_core": "rad_core0",
            "sld_solvent": "sld_solv",
            "thickness": "thick_shell",
            "sld_in": "sld_in_shell",
            "sld_out": "sld_out_shell"
        }
    ],
    "parallelepiped": [
        "ParallelepipedModel",
        {
            "phi": "parallel_phi",
            "psi": "parallel_psi",
            "sld_solvent": "sldSolv",
            "length_a": "short_a",
            "length_b": "short_b",
            "sld": "sldPipe",
            "theta": "parallel_theta",
            "length_c": "long_c",
            "M0:sld": "M0_sld_pipe",
            "mtheta:sld": "M_theta_pipe",
            "mphi:sld": "M_phi_pipe",
            "M0:sld_solvent": "M0_sld_solv",
            "mtheta:sld_solvent": "M_theta_solv",
            "mphi:sld_solvent": "M_phi_solv",
            "up:frac_i": "Up_frac_i",
            "up:frac_f": "Up_frac_f",
            "up:angle": "Up_theta",
        }
    ],
    "peak_lorentz": [
        "Peak Lorentz Model",
        {
            "peak_pos": "q0",
            "peak_hwhm": "B"
        },
        "PeakLorentzModel",
    ],
    "pearl_necklace": [
        "PearlNecklaceModel",
        {
            "scale": "scale",
            "thick_string": "thick_string",
            "sld_string": "sld_string",
            "sld_solvent": "sld_solv",
            "edge_sep": "edge_separation",
            "num_pearls": "num_pearls",
            "radius": "radius",
            "background": "background",
            "sld": "sld_pearl"
        }
    ],
    "poly_gauss_coil": [
        "Poly_GaussCoil",
        {
            "rg": "rg",
            "polydispersity": "poly_m",
            "i_zero": "scale",
            "background": "background",
        }
    ],
    "polymer_excl_volume": [
        "PolymerExclVolume",
        {
            "rg": "rg",
            "scale": "scale",
            "background": "background",
            "porod_exp": "m"
        }
    ],
    "polymer_micelle": [
        "MicelleSphCoreModel",
        {
            "sld_corona": "rho_corona",
            "sld_solvent": "rho_solv",
            "sld_core": "rho_core",
            "ndensity": "ndensity",
            "v_core": "v_core",
            "v_corona": "v_corona",
            "radius_core": "radius_core",
            "rg": "radius_gyr",
            "d_penetration": "d_penetration",
            "n_aggreg": "n_aggreg",
        }
    ],
    "porod": [
        "PorodModel",
        {
            "scale": "scale",
            "background": "background"
        }
    ],
    "power_law": [
        "PowerLawAbsModel",
        {
            "scale": "scale",
            "background": "background",
            "power": "m"
        }
    ],
    "pringle": [
        "PringlesModel",
        {
            "scale": "scale",
            "sld_solvent": "sld_solvent",
            "thickness": "thickness",
            "beta": "beta",
            "radius": "radius",
            "background": "background",
            "alpha": "alpha",
            "sld": "sld_pringle"
        }
    ],
    "raspberry": [
        "RaspBerryModel",
        {
            "volfraction_lg": "volf_Lsph",
            "volfraction_sm": "volf_Ssph",
            "radius_sm": "radius_Ssph",
            "radius_lg": "radius_Lsph",
            "sld_lg": "sld_Lsph",
            "sld_sm": "sld_Ssph",
            "sld_solvent": "sld_solv",
            "surface_fraction": "surfrac_Ssph",
            "penetration": "delta_Ssph"
        }
    ],
    "rectangular_prism": [
        "RectangularPrismModel",
        {
            "sld": "sldPipe",
            "length_a": "short_side",
            "b2a_ratio": "b2a_ratio",
            "c2a_ratio": "c2a_ratio",
            "sld_solvent": "sldSolv"
        }
    ],
    "rpa": [
        "RPA10Model",
        {
            "K12": "Kab", "K13": "Kac", "K14": "Kad",
            "K23": "Kbc", "K24": "Kbd", "K34": "Kcd",
            "N1": "Na", "N2": "Nb", "N3": "Nc", "N4": "Nd",
            "L1": "La", "L2": "Lb", "L3": "Lc", "L4": "Ld",
            "v1": "va", "v2": "vb", "v3": "vc", "v4": "vd",
            "b1": "ba", "b2": "bb", "b3": "bc", "b4": "bd",
            "Phi1": "Phia", "Phi2": "Phib", "Phi3": "Phic", "Phi4": "Phid",
            "case_num": "lcase_n"
        }
    ],
    "sc_paracrystal": [
        "SCCrystalModel",
        {
            "sld": "sldSph",
            "sld_solvent": "sldSolv"
        }
    ],
    "sphere": [
        "SphereModel",
        {
            "sld": "sldSph",
            "radius": "radius",
            "sld_solvent": "sldSolv",
            "M0:sld": "M0_sld_sph",
            "mtheta:sld": "M_theta_sph",
            "mphi:sld": "M_phi_sph",
            "M0:sld_solvent": "M0_sld_solv",
            "mtheta:sld_solvent": "M_theta_solv",
            "mphi:sld_solvent": "M_phi_solv",
            "up:frac_i": "Up_frac_i",
            "up:frac_f": "Up_frac_f",
            "up:angle": "Up_theta"
        }
    ],
    "spherical_sld": [
        "SphericalSLDModel",
        # Be lazy and use a generator expression to define
        #    sld1: sld_flat0, ...
        #    thickness1: thick_flat0, ...
        #    interface1: thick_inter0, ...
        #    shape1: func_inter0, ...
        #    nu1: nu_inter0, ...
        # but override thickness1 => rad_cor0 and sld1 => sld_core0.
        # Note: explicit key,value pairs given by **{...} override the
        # keys from the gnerator expression ((k,v) for k,v in seq) when
        # used as dict((generator), **{...})
        dict(((field_new+str(index+1), field_old+str(index))
              for field_new, field_old in [("sld", "sld_flat"),
                                           ("thickness", "thick_flat"),
                                           ("interface", "thick_inter"),
                                           ("shape", "func_inter"),
                                           ("nu", "nu_inter"),]
              for index in range(11)),
             **{
                   "n_shells": "n_shells",
                   "n_steps": "npts_inter",
                   "sld_solvent": "sld_solv",
                   "thickness1": "rad_core0",
                   "sld1": "sld_core0",
               })
    ],
    "squarewell": [
        "SquareWellStructure",
        {
            "scale": "scale_factor",
            "radius_effective": "effect_radius",
            "wellwidth": "wellwidth",
            "welldepth": "welldepth",
        }
    ],
    "stacked_disks": [
        "StackedDisksModel",
        {
            "phi": "axis_phi",
            "sld_layer": "layer_sld",
            "sld_core": "core_sld",
            "theta": "axis_theta",
            "sld_solvent": "solvent_sld",
            "n_stacking": "n_stacking",
            "thick_layer": "layer_thick",
            "thick_core": "core_thick",
        }
    ],
    "star_polymer": [
        "StarPolymer",
        {
            "arms": "arms",
            "rg_squared": "R2"
        }
    ],
    "stickyhardsphere": [
        "StickyHSStructure",
        {
            "scale": "scale_factor",
            "radius_effective": "effect_radius",
        }
    ],
    "surface_fractal": [
        "SurfaceFractalModel",
        {
            "cutoff_length": "co_length",
            "radius": "radius",
            "fractal_dim_surf": "surface_dim"
        }
    ],
    "teubner_strey": [
        "TeubnerStreyModel",
        {
            # Note: parameters are completely rewritten in convert.py
            "volfraction_a": "volfraction_a",
            "sld_a": "sld_a",
            "sld_b": "sld_b",
            "d": "d",
            "xi": "xi",
        }
    ],
    "triaxial_ellipsoid": [
        "TriaxialEllipsoidModel",
        {
            "phi": "axis_phi",
            "radius_equat_minor": "semi_axisA",
            "radius_polar": "semi_axisC",
            "radius_equat_major": "semi_axisB",
            "sld_solvent": "sldSolv",
            "psi": "axis_psi",
            "sld": "sldEll",
            "theta": "axis_theta"
        }
    ],
    "two_lorentzian": [
        "TwoLorentzian",
        {
            "lorentz_scale_1": "scale_1",
            "lorentz_scale_2": "scale_2",
            "lorentz_exp_1": "exponent_1",
            "lorentz_exp_2": "exponent_2",
            "lorentz_length_2": "length_2",
            "lorentz_length_1": "length_1",
            "background": "background"
        },
        "TwoLorentzianModel",
    ],
    "two_power_law": [
        "TwoPowerLaw",
        {
            "coefficent_1": "coef_A",
            "power_2": "power2",
            "power_1": "power1",
            "background": "background",
            "crossover": "qc"
        },
        "TwoPowerLawModel",
    ],
    "unified_power_Rg": [
        "UnifiedPowerRg",
        dict(((field_new+str(index), field_old+str(index))
              for field_new, field_old in [("rg", "Rg"),
                                           ("power", "power"),
                                           ("G", "G"),
                                           ("B", "B"),]
              for index in range(11)),
             **{
                   "background": "background",
                   "scale": "scale",
               }),
        "UnifiedPowerRgModel",
    ],
    "vesicle": [
        "VesicleModel",
        {
            "sld": "shell_sld",
            "sld_solvent": "solv_sld"
        }
    ]
    }
}



"""
Core model handling routines.
"""
from __future__ import print_function

__all__ = [
    "list_models", "load_model", "load_model_info",
    "build_model", "precompile_dlls",
    ]

import os
import re
from os.path import basename, dirname, join as joinpath
from glob import glob

import numpy as np # type: ignore

from . import generate
from . import modelinfo
from . import product
from . import mixture
from . import kernelpy
from . import kerneldll
from . import custom

if os.environ.get("SAS_OPENCL", "").lower() == "none":
    HAVE_OPENCL = False
else:
    try:
        from . import kernelcl
        HAVE_OPENCL = True
    except Exception:
        HAVE_OPENCL = False

CUSTOM_MODEL_PATH = os.environ.get('SAS_MODELPATH', "")
if CUSTOM_MODEL_PATH == "":
    path = joinpath(os.path.expanduser("~"), ".sasmodels", "custom_models")
    if not os.path.isdir(path):
        os.makedirs(path)
    CUSTOM_MODEL_PATH = path

try:
    from typing import List, Union, Optional, Any
    from .kernel import KernelModel
    from .modelinfo import ModelInfo
except ImportError:
    pass

# TODO: refactor composite model support
# The current load_model_info/build_model does not reuse existing model
# definitions when loading a composite model, instead reloading and
# rebuilding the kernel for each component model in the expression.  This
# is fine in a scripting environment where the model is built when the script
# starts and is thrown away when the script ends, but may not be the best
# solution in a long-lived application.  This affects the following functions:
#
#    load_model
#    load_model_info
#    build_model

KINDS = ("all", "py", "c", "double", "single", "opencl", "1d", "2d",
         "nonmagnetic", "magnetic")
def list_models(kind=None):
    # type: (str) -> List[str]
    """
    Return the list of available models on the model path.

    *kind* can be one of the following:

        * all: all models
        * py: python models only
        * c: compiled models only
        * single: models which support single precision
        * double: models which require double precision
        * opencl: controls if OpenCL is supperessed
        * 1d: models which are 1D only, or 2D using abs(q)
        * 2d: models which can be 2D
        * magnetic: models with an sld
        * nommagnetic: models without an sld

    For multiple conditions, combine with plus.  For example, *c+single+2d*
    would return all oriented models implemented in C which can be computed
    accurately with single precision arithmetic.
    """
    if kind and any(k not in KINDS for k in kind.split('+')):
        raise ValueError("kind not in " + ", ".join(KINDS))
    files = sorted(glob(joinpath(generate.MODEL_PATH, "[a-zA-Z]*.py")))
    available_models = [basename(f)[:-3] for f in files]
    if kind and '+' in kind:
        all_kinds = kind.split('+')
        condition = lambda name: all(_matches(name, k) for k in all_kinds)
    else:
        condition = lambda name: _matches(name, kind)
    selected = [name for name in available_models if condition(name)]

    return selected

def _matches(name, kind):
    if kind is None or kind == "all":
        return True
    info = load_model_info(name)
    pars = info.parameters.kernel_parameters
    if kind == "py" and callable(info.Iq):
        return True
    elif kind == "c" and not callable(info.Iq):
        return True
    elif kind == "double" and not info.single:
        return True
    elif kind == "single" and info.single:
        return True
    elif kind == "opencl" and info.opencl:
        return True
    elif kind == "2d" and any(p.type == 'orientation' for p in pars):
        return True
    elif kind == "1d" and all(p.type != 'orientation' for p in pars):
        return True
    elif kind == "magnetic" and any(p.type == 'sld' for p in pars):
        return True
    elif kind == "nonmagnetic" and any(p.type != 'sld' for p in pars):
        return True
    return False

def load_model(model_name, dtype=None, platform='ocl'):
    # type: (str, str, str) -> KernelModel
    """
    Load model info and build model.

    *model_name* is the name of the model, or perhaps a model expression
    such as sphere*hardsphere or sphere+cylinder.

    *dtype* and *platform* are given by :func:`build_model`.
    """
    return build_model(load_model_info(model_name),
                       dtype=dtype, platform=platform)

def load_model_info(model_string):
    # type: (str) -> modelinfo.ModelInfo
    """
    Load a model definition given the model name.

    *model_string* is the name of the model, or perhaps a model expression
    such as sphere*cylinder or sphere+cylinder. Use '@' for a structure
    factor product, e.g. sphere@hardsphere. Custom models can be specified by
    prefixing the model name with 'custom.', e.g. 'custom.MyModel+sphere'.

    This returns a handle to the module defining the model.  This can be
    used with functions in generate to build the docs or extract model info.
    """
    if '@' in model_string:
        parts = model_string.split('@')
        if len(parts) != 2:
            raise ValueError("Use P@S to apply a structure factor S to model P")
        P_info, Q_info = [load_model_info(part) for part in parts]
        return product.make_product_info(P_info, Q_info)

    product_parts = []
    addition_parts = []

    addition_parts_names = model_string.split('+')
    if len(addition_parts_names) >= 2:
        addition_parts = [load_model_info(part) for part in addition_parts_names]
    elif len(addition_parts_names) == 1:
        product_parts_names = model_string.split('*')
        if len(product_parts_names) >= 2:
            product_parts = [load_model_info(part) for part in product_parts_names]
        elif len(product_parts_names) == 1:
            if "custom." in product_parts_names[0]:
                # Extract ModelName from "custom.ModelName"
                pattern = "custom.([A-Za-z0-9_-]+)"
                result = re.match(pattern, product_parts_names[0])
                if result is None:
                    raise ValueError("Model name in invalid format: " + product_parts_names[0])
                model_name = result.group(1)
                # Use ModelName to find the path to the custom model file
                model_path = joinpath(CUSTOM_MODEL_PATH, model_name + ".py")
                if not os.path.isfile(model_path):
                    raise ValueError("The model file {} doesn't exist".format(model_path))
                kernel_module = custom.load_custom_kernel_module(model_path)
                return modelinfo.make_model_info(kernel_module)
            # Model is a core model
            kernel_module = generate.load_kernel_module(product_parts_names[0])
            return modelinfo.make_model_info(kernel_module)

    model = None
    if len(product_parts) > 1:
        model = mixture.make_mixture_info(product_parts, operation='*')
    if len(addition_parts) > 1:
        if model is not None:
            addition_parts.append(model)
        model = mixture.make_mixture_info(addition_parts, operation='+')
    return model


def build_model(model_info, dtype=None, platform="ocl"):
    # type: (modelinfo.ModelInfo, str, str) -> KernelModel
    """
    Prepare the model for the default execution platform.

    This will return an OpenCL model, a DLL model or a python model depending
    on the model and the computing platform.

    *model_info* is the model definition structure returned from
    :func:`load_model_info`.

    *dtype* indicates whether the model should use single or double precision
    for the calculation.  Choices are 'single', 'double', 'quad', 'half',
    or 'fast'.  If *dtype* ends with '!', then force the use of the DLL rather
    than OpenCL for the calculation.

    *platform* should be "dll" to force the dll to be used for C models,
    otherwise it uses the default "ocl".
    """
    composition = model_info.composition
    if composition is not None:
        composition_type, parts = composition
        models = [build_model(p, dtype=dtype, platform=platform) for p in parts]
        if composition_type == 'mixture':
            return mixture.MixtureModel(model_info, models)
        elif composition_type == 'product':
            P, S = models
            return product.ProductModel(model_info, P, S)
        else:
            raise ValueError('unknown mixture type %s'%composition_type)

    # If it is a python model, return it immediately
    if callable(model_info.Iq):
        return kernelpy.PyModel(model_info)

    numpy_dtype, fast, platform = parse_dtype(model_info, dtype, platform)

    source = generate.make_source(model_info)
    if platform == "dll":
        #print("building dll", numpy_dtype)
        return kerneldll.load_dll(source['dll'], model_info, numpy_dtype)
    else:
        #print("building ocl", numpy_dtype)
        return kernelcl.GpuModel(source, model_info, numpy_dtype, fast=fast)

def precompile_dlls(path, dtype="double"):
    # type: (str, str) -> List[str]
    """
    Precompile the dlls for all builtin models, returning a list of dll paths.

    *path* is the directory in which to save the dlls.  It will be created if
    it does not already exist.

    This can be used when build the windows distribution of sasmodels
    which may be missing the OpenCL driver and the dll compiler.
    """
    numpy_dtype = np.dtype(dtype)
    if not os.path.exists(path):
        os.makedirs(path)
    compiled_dlls = []
    for model_name in list_models():
        model_info = load_model_info(model_name)
        if not callable(model_info.Iq):
            source = generate.make_source(model_info)['dll']
            old_path = kerneldll.DLL_PATH
            try:
                kerneldll.DLL_PATH = path
                dll = kerneldll.make_dll(source, model_info, dtype=numpy_dtype)
            finally:
                kerneldll.DLL_PATH = old_path
            compiled_dlls.append(dll)
    return compiled_dlls

def parse_dtype(model_info, dtype=None, platform=None):
    # type: (ModelInfo, str, str) -> (np.dtype, bool, str)
    """
    Interpret dtype string, returning np.dtype and fast flag.

    Possible types include 'half', 'single', 'double' and 'quad'.  If the
    type is 'fast', then this is equivalent to dtype 'single' but using
    fast native functions rather than those with the precision level guaranteed
    by the OpenCL standard.

    Platform preference can be specfied ("ocl" vs "dll"), with the default
    being OpenCL if it is availabe.  If the dtype name ends with '!' then
    platform is forced to be DLL rather than OpenCL.

    This routine ignores the preferences within the model definition.  This
    is by design.  It allows us to test models in single precision even when
    we have flagged them as requiring double precision so we can easily check
    the performance on different platforms without having to change the model
    definition.
    """
    # Assign default platform, overriding ocl with dll if OpenCL is unavailable
    # If opencl=False OpenCL is switched off

    if platform is None:
        platform = "ocl"
    if platform == "ocl" and not HAVE_OPENCL or not model_info.opencl:
        platform = "dll"

    # Check if type indicates dll regardless of which platform is given
    if dtype is not None and dtype.endswith('!'):
        platform = "dll"
        dtype = dtype[:-1]

    # Convert special type names "half", "fast", and "quad"
    fast = (dtype == "fast")
    if fast:
        dtype = "single"
    elif dtype == "quad":
        dtype = "longdouble"
    elif dtype == "half":
        dtype = "float16"

    # Convert dtype string to numpy dtype.
    if dtype is None or dtype == "default":
        numpy_dtype = (generate.F32 if platform == "ocl" and model_info.single
                       else generate.F64)
    else:
        numpy_dtype = np.dtype(dtype)

    # Make sure that the type is supported by opencl, otherwise use dll
    if platform == "ocl":
        env = kernelcl.environment()
        if not env.has_type(numpy_dtype):
            platform = "dll"
            if dtype is None:
                numpy_dtype = generate.F64

    return numpy_dtype, fast, platform

def list_models_main():
    # type: () -> None
    """
    Run list_models as a main program.  See :func:`list_models` for the
    kinds of models that can be requested on the command line.
    """
    import sys
    kind = sys.argv[1] if len(sys.argv) > 1 else "all"
    print("\n".join(list_models(kind)))

if __name__ == "__main__":
    list_models_main()



r"""
This model calculates intensity using simple linear function

Definition
----------

The scattering intensity $I(q)$ is calculated as

.. math::

    I(q) = \text{scale} (A + B \cdot q) + \text{background}

.. note::
    For 2D plots intensity has different definition than other shape independent models

.. math::

    I(q) = \text{scale} (I(qx) \cdot I(qy)) + \text{background}

References
----------

None.

"""
from numpy import inf

name = "line"
title = "Line model"
description = """\
      I(q) = A + B*q

      List of default parameters:
      A = intercept
      B = slope
      """
category = "shape-independent"

# pylint: disable=bad-whitespace, line-too-long
#             ["name", "units", default, [lower, upper], "type", "description"],
parameters = [["intercept", "1/cm",   1.0, [-inf, inf], "", "intercept in linear model"],
              ["slope",     "Ang/cm", 1.0, [-inf, inf], "", "slope in linear model"],
             ]
# pylint: enable=bad-whitespace, line-too-long

def Iq(q, intercept, slope):
    """
    :param q:           Input q-value
    :param intercept:   Intrecept in linear model
    :param slope:       Slope in linear model
    :return:            Calculated Intensity
    """
    inten = intercept + slope*q
    return inten

Iq.vectorized = True # Iq accepts an array of q values

def Iqxy(qx, qy, *args):
    """
    :param qx:   Input q_x-value
    :param qy:   Input q_y-value
    :param args: Remaining arguments
    :return:     2D-Intensity
    """
    # TODO: SasView documents 2D intensity as Iq(qx)*Iq(qy), but returns Iq(qy)
    # Note: SasView.run([r, theta]) does return Iq(qx)*Iq(qy)
    return Iq(qx, *args)*Iq(qy, *args)

Iqxy.vectorized = True  # Iqxy accepts an array of qx qy values

def random():
    import numpy as np
    scale = 10**np.random.uniform(0, 3)
    slope = np.random.uniform(-1, 1)*1e2
    offset = 1e-5 + (0 if slope > 0 else -slope)
    intercept = 10**np.random.uniform(0, 1) + offset
    pars = dict(
        #background=0,
        scale=scale,
        slope=slope,
        intercept=intercept,
    )
    return pars

tests = [
    [{'intercept': 1.0, 'slope': 1.0, }, 1.0, 2.001],
    [{'intercept': 1.0, 'slope': 1.0, }, 0.0, 1.001],
    [{'intercept': 1.0, 'slope': 1.0, }, 0.4, 1.401],
    [{'intercept': 1.0, 'slope': 1.0, }, 1.3, 2.301],
    [{'intercept': 1.0, 'slope': 1.0, }, 0.5, 1.501],
    [{'intercept': 1.0, 'slope': 1.0, }, [0.4, 0.5], [1.401, 1.501]],
    [{'intercept': 1.0, 'slope': 1.0, 'background': 0.0, }, (1.3, 1.57), 5.911],
]



from bumps.names import *

from sasmodels.data import load_data
from sasmodels.core import load_model
from sasmodels.bumps_model import Model, Experiment

# Spherical particle data, not ellipsoids
sans, usans = load_data('latex_smeared.xml', index='all')
usans.qmin, usans.qmax = np.min(usans.x), np.max(usans.x)
usans.mask = (usans.x < 0.0)
usans.oriented = True
#print sans.dxl, usans.dxl
#import pprint; pprint.pprint(sans.__dict__)

kernel = load_model("ellipsoid")

model = Model(
    kernel,
    scale=0.08, background=0,
    sld=.291, sld_solvent=7.105,
    radius_polar=1800, radius_polar_pd=0.222296, radius_polar_pd_n=0,
    radius_equatorial=2600, radius_equatorial_pd=0.28, radius_equatorial_pd_n=0,
    theta=60, theta_pd=0, theta_pd_n=0,
    phi=60, phi_pd=0, phi_pd_n=0,
    )

# SET THE FITTING PARAMETERS
model.radius_polar.range(1000, 10000)
model.radius_equatorial.range(1000, 10000)
model.theta.range(0, 360)
model.phi.range(0, 360)
model.background.range(0,1000)
model.scale.range(0, 10)


#cutoff = 0     # no cutoff on polydisperisity loops
#cutoff = 1e-5  # default cutoff
cutoff = 1e-3  # low precision cutoff
M = Experiment(data=usans, model=model, cutoff=cutoff)
problem = FitProblem(M)




"""
SAS model constructor.

Small angle scattering models are defined by a set of kernel functions:

    *Iq(q, p1, p2, ...)* returns the scattering at q for a form with
    particular dimensions averaged over all orientations.

    *Iqxy(qx, qy, p1, p2, ...)* returns the scattering at qx, qy for a form
    with particular dimensions for a single orientation.

    *Imagnetic(qx, qy, result[], p1, p2, ...)* returns the scattering for the
    polarized neutron spin states (up-up, up-down, down-up, down-down) for
    a form with particular dimensions for a single orientation.

    *form_volume(p1, p2, ...)* returns the volume of the form with particular
    dimension, or 1.0 if no volume normalization is required.

    *ER(p1, p2, ...)* returns the effective radius of the form with
    particular dimensions.

    *VR(p1, p2, ...)* returns the volume ratio for core-shell style forms.

    #define INVALID(v) (expr)  returns False if v.parameter is invalid
    for some parameter or other (e.g., v.bell_radius < v.radius).  If
    necessary, the expression can call a function.

These functions are defined in a kernel module .py script and an associated
set of .c files.  The model constructor will use them to create models with
polydispersity across volume and orientation parameters, and provide
scale and background parameters for each model.

*Iq*, *Iqxy*, *Imagnetic* and *form_volume* should be stylized C-99
functions written for OpenCL.  All functions need prototype declarations
even if the are defined before they are used.  OpenCL does not support
*#include* preprocessor directives, so instead the list of includes needs
to be given as part of the metadata in the kernel module definition.
The included files should be listed using a path relative to the kernel
module, or if using "lib/file.c" if it is one of the standard includes
provided with the sasmodels source.  The includes need to be listed in
order so that functions are defined before they are used.

Floating point values should be declared as *double*.  For single precision
calculations, *double* will be replaced by *float*.  The single precision
conversion will also tag floating point constants with "f" to make them
single precision constants.  When using integral values in floating point
expressions, they should be expressed as floating point values by including
a decimal point.  This includes 0., 1. and 2.

OpenCL has a *sincos* function which can improve performance when both
the *sin* and *cos* values are needed for a particular argument.  Since
this function does not exist in C99, all use of *sincos* should be
replaced by the macro *SINCOS(value, sn, cn)* where *sn* and *cn* are
previously declared *double* variables.  When compiled for systems without
OpenCL, *SINCOS* will be replaced by *sin* and *cos* calls.   If *value* is
an expression, it will appear twice in this case; whether or not it will be
evaluated twice depends on the quality of the compiler.

If the input parameters are invalid, the scattering calculator should
return a negative number. Particularly with polydispersity, there are
some sets of shape parameters which lead to nonsensical forms, such
as a capped cylinder where the cap radius is smaller than the
cylinder radius.  The polydispersity calculation will ignore these points,
effectively chopping the parameter weight distributions at the boundary
of the infeasible region.  The resulting scattering will be set to
background.  This will work correctly even when polydispersity is off.

*ER* and *VR* are python functions which operate on parameter vectors.
The constructor code will generate the necessary vectors for computing
them with the desired polydispersity.
The kernel module must set variables defining the kernel meta data:

    *id* is an implicit variable formed from the filename.  It will be
    a valid python identifier, and will be used as the reference into
    the html documentation, with '_' replaced by '-'.

    *name* is the model name as displayed to the user.  If it is missing,
    it will be constructed from the id.

    *title* is a short description of the model, suitable for a tool tip,
    or a one line model summary in a table of models.

    *description* is an extended description of the model to be displayed
    while the model parameters are being edited.

    *parameters* is the list of parameters.  Parameters in the kernel
    functions must appear in the same order as they appear in the
    parameters list.  Two additional parameters, *scale* and *background*
    are added to the beginning of the parameter list.  They will show up
    in the documentation as model parameters, but they are never sent to
    the kernel functions.  Note that *effect_radius* and *volfraction*
    must occur first in structure factor calculations.

    *category* is the default category for the model.  The category is
    two level structure, with the form "group:section", indicating where
    in the manual the model will be located.  Models are alphabetical
    within their section.

    *source* is the list of C-99 source files that must be joined to
    create the OpenCL kernel functions.  The files defining the functions
    need to be listed before the files which use the functions.

    *ER* is a python function defining the effective radius.  If it is
    not present, the effective radius is 0.

    *VR* is a python function defining the volume ratio.  If it is not
    present, the volume ratio is 1.

    *form_volume*, *Iq*, *Iqxy*, *Imagnetic* are strings containing the
    C source code for the body of the volume, Iq, and Iqxy functions
    respectively.  These can also be defined in the last source file.

    *Iq* and *Iqxy* also be instead be python functions defining the
    kernel.  If they are marked as *Iq.vectorized = True* then the
    kernel is passed the entire *q* vector at once, otherwise it is
    passed values one *q* at a time.  The performance improvement of
    this step is significant.

    *demo* is a dictionary of parameter=value defining a set of
    parameters to use by default when *compare* is called.  Any
    parameter not set in *demo* gets the initial value from the
    parameter list.  *demo* is mostly needed to set the default
    polydispersity values for tests.

A :class:`modelinfo.ModelInfo` structure is constructed from the kernel meta
data and returned to the caller.

The doc string at the start of the kernel module will be used to
construct the model documentation web pages.  Embedded figures should
appear in the subdirectory "img" beside the model definition, and tagged
with the kernel module name to avoid collision with other models.  Some
file systems are case-sensitive, so only use lower case characters for
file names and extensions.

Code follows the C99 standard with the following extensions and conditions::

    M_PI_180 = pi/180
    M_4PI_3 = 4pi/3
    square(x) = x*x
    cube(x) = x*x*x
    sas_sinx_x(x) = sin(x)/x, with sin(0)/0 -> 1
    all double precision constants must include the decimal point
    all double declarations may be converted to half, float, or long double
    FLOAT_SIZE is the number of bytes in the converted variables

:func:`load_kernel_module` loads the model definition file and
:func:`modelinfo.make_model_info` parses it. :func:`make_source`
converts C-based model definitions to C source code, including the
polydispersity integral.  :func:`model_sources` returns the list of
source files the model depends on, and :func:`timestamp` returns
the latest time stamp amongst the source files (so you can check if
the model needs to be rebuilt).

The function :func:`make_doc` extracts the doc string and adds the
parameter table to the top.  *make_figure* in *sasmodels/doc/genmodel*
creates the default figure for the model.  [These two sets of code
should mignrate into docs.py so docs can be updated in one place].
"""
from __future__ import print_function

# TODO: determine which functions are useful outside of generate
#__all__ = ["model_info", "make_doc", "make_source", "convert_type"]

import sys
from os.path import abspath, dirname, join as joinpath, exists, isdir, getmtime
import re
import string
from zlib import crc32

import numpy as np  # type: ignore

from .modelinfo import Parameter
from .custom import load_custom_kernel_module

try:
    from typing import Tuple, Sequence, Iterator, Dict
    from .modelinfo import ModelInfo
except ImportError:
    pass

def get_data_path(external_dir, target_file):
    path = abspath(dirname(__file__))
    if exists(joinpath(path, target_file)):
        return path

    # check next to exe/zip file
    exepath = dirname(sys.executable)
    path = joinpath(exepath, external_dir)
    if exists(joinpath(path, target_file)):
        return path

    # check in py2app Contents/Resources
    path = joinpath(exepath, '..', 'Resources', external_dir)
    if exists(joinpath(path, target_file)):
        return abspath(path)

    raise RuntimeError('Could not find '+joinpath(external_dir, target_file))

EXTERNAL_DIR = 'sasmodels-data'
DATA_PATH = get_data_path(EXTERNAL_DIR, 'kernel_template.c')
MODEL_PATH = joinpath(DATA_PATH, 'models')

F16 = np.dtype('float16')
F32 = np.dtype('float32')
F64 = np.dtype('float64')
try:  # CRUFT: older numpy does not support float128
    F128 = np.dtype('float128')
except TypeError:
    F128 = None

# Conversion from units defined in the parameter table for each model
# to units displayed in the sphinx documentation.
# This section associates the unit with the macro to use to produce the LaTex
# code.  The macro itself needs to be defined in sasmodels/doc/rst_prolog.
#
# NOTE: there is an RST_PROLOG at the end of this file which is NOT
# used for the bundled documentation. Still as long as we are defining the macros
# in two places any new addition should define the macro in both places.
RST_UNITS = {
    "Ang": "|Ang|",
    "1/Ang": "|Ang^-1|",
    "1/Ang^2": "|Ang^-2|",
    "Ang^3": "|Ang^3|",
    "Ang^2": "|Ang^2|",
    "1e15/cm^3": "|1e15cm^3|",
    "Ang^3/mol": "|Ang^3|/mol",
    "1e-6/Ang^2": "|1e-6Ang^-2|",
    "degrees": "degree",
    "1/cm": "|cm^-1|",
    "Ang/cm": "|Ang*cm^-1|",
    "g/cm^3": "|g/cm^3|",
    "mg/m^2": "|mg/m^2|",
    "": "None",
    }

# Headers for the parameters tables in th sphinx documentation
PARTABLE_HEADERS = [
    "Parameter",
    "Description",
    "Units",
    "Default value",
    ]

# Minimum width for a default value (this is shorter than the column header
# width, so will be ignored).
PARTABLE_VALUE_WIDTH = 10

# Documentation header for the module, giving the model name, its short
# description and its parameter table.  The remainder of the doc comes
# from the module docstring.
DOC_HEADER = """.. _%(id)s:

%(name)s
=======================================================

%(title)s

%(parameters)s

%(returns)s

%(docs)s
"""


def format_units(units):
    # type: (str) -> str
    """
    Convert units into ReStructured Text format.
    """
    return "string" if isinstance(units, list) else RST_UNITS.get(units, units)


def make_partable(pars):
    # type: (List[Parameter]) -> str
    """
    Generate the parameter table to include in the sphinx documentation.
    """
    column_widths = [
        max(len(p.name) for p in pars),
        max(len(p.description) for p in pars),
        max(len(format_units(p.units)) for p in pars),
        PARTABLE_VALUE_WIDTH,
        ]
    column_widths = [max(w, len(h))
                     for w, h in zip(column_widths, PARTABLE_HEADERS)]

    sep = " ".join("="*w for w in column_widths)
    lines = [
        sep,
        " ".join("%-*s" % (w, h)
                 for w, h in zip(column_widths, PARTABLE_HEADERS)),
        sep,
        ]
    for p in pars:
        lines.append(" ".join([
            "%-*s" % (column_widths[0], p.name),
            "%-*s" % (column_widths[1], p.description),
            "%-*s" % (column_widths[2], format_units(p.units)),
            "%*g" % (column_widths[3], p.default),
            ]))
    lines.append(sep)
    return "\n".join(lines)


def _search(search_path, filename):
    # type: (List[str], str) -> str
    """
    Find *filename* in *search_path*.

    Raises ValueError if file does not exist.
    """
    for path in search_path:
        target = joinpath(path, filename)
        if exists(target):
            return target
    raise ValueError("%r not found in %s" % (filename, search_path))


def model_sources(model_info):
    # type: (ModelInfo) -> List[str]
    """
    Return a list of the sources file paths for the module.
    """
    search_path = [dirname(model_info.filename), MODEL_PATH]
    return [_search(search_path, f) for f in model_info.source]


def dll_timestamp(model_info):
    # type: (ModelInfo) -> int
    """
    Return a timestamp for the model corresponding to the most recently
    changed file or dependency.
    """
    # TODO: fails DRY; templates appear two places.
    model_templates = [joinpath(DATA_PATH, filename)
                       for filename in ('kernel_header.c', 'kernel_iq.c')]
    source_files = (model_sources(model_info)
                    + model_templates
                    + [model_info.filename])
    # Note: file may not exist when it is a standard model from library.zip
    times = [getmtime(f) for f in source_files if exists(f)]
    newest = max(times) if times else 0
    return newest

def ocl_timestamp(model_info):
    # type: (ModelInfo) -> int
    """
    Return a timestamp for the model corresponding to the most recently
    changed file or dependency.

    Note that this does not look at the time stamps for the OpenCL header
    information since that need not trigger a recompile of the DLL.
    """
    # TODO: fails DRY; templates appear two places.
    model_templates = [joinpath(DATA_PATH, filename)
                       for filename in ('kernel_header.c', 'kernel_iq.cl')]
    source_files = (model_sources(model_info)
                    + model_templates
                    + [model_info.filename])
    # Note: file may not exist when it is a standard model from library.zip
    times = [getmtime(f) for f in source_files if exists(f)]
    newest = max(times) if times else 0
    return newest

def tag_source(source):
    # type: (str) -> str
    """
    Return a unique tag for the source code.
    """
    # Note: need 0xffffffff&val to force an unsigned 32-bit number
    try:
        source = source.encode('utf8')
    except AttributeError: # bytes has no encode attribute in python 3
        pass
    return "%08X"%(0xffffffff&crc32(source))

def convert_type(source, dtype):
    # type: (str, np.dtype) -> str
    """
    Convert code from double precision to the desired type.

    Floating point constants are tagged with 'f' for single precision or 'L'
    for long double precision.
    """
    source = _fix_tgmath_int(source)
    if dtype == F16:
        fbytes = 2
        source = _convert_type(source, "half", "f")
    elif dtype == F32:
        fbytes = 4
        source = _convert_type(source, "float", "f")
    elif dtype == F64:
        fbytes = 8
        # no need to convert if it is already double
    elif dtype == F128:
        fbytes = 16
        source = _convert_type(source, "long double", "L")
    else:
        raise ValueError("Unexpected dtype in source conversion: %s" % dtype)
    return ("#define FLOAT_SIZE %d\n" % fbytes)+source


def _convert_type(source, type_name, constant_flag):
    # type: (str, str, str) -> str
    """
    Replace 'double' with *type_name* in *source*, tagging floating point
    constants with *constant_flag*.
    """
    # Convert double keyword to float/long double/half.
    # Accept an 'n' # parameter for vector # values, where n is 2, 4, 8 or 16.
    # Assume complex numbers are represented as cdouble which is typedef'd
    # to double2.
    source = re.sub(r'(^|[^a-zA-Z0-9_]c?)double(([248]|16)?($|[^a-zA-Z0-9_]))',
                    r'\1%s\2'%type_name, source)
    source = _tag_float(source, constant_flag)
    return source

TGMATH_INT_RE = re.compile(r"""
(?: # Non-capturing match; not lookbehind since pattern length is variable
  \b              # word boundary
   # various math functions
  (a?(sin|cos|tan)h? | atan2
   | erfc? | tgamma
   | exp(2|10|m1)? | log(2|10|1p)? | pow[nr]? | sqrt | rsqrt | rootn
   | fabs | fmax | fmin
   )
  \s*[(]\s*       # open parenthesis
)
[+-]?(0|[1-9]\d*) # integer
(?=               # lookahead match: don't want to move from end of int
  \s*[,)]         # comma or close parenthesis for end of argument
)                 # end lookahead
""", re.VERBOSE)
def _fix_tgmath_int(source):
    # type: (str) -> str
    """
    Replace f(integer) with f(integer.) for sin, cos, pow, etc.

    OS X OpenCL complains that it can't resolve the type generic calls to
    the standard math functions when they are called with integer constants,
    but this does not happen with the Windows Intel driver for example.
    To avoid confusion on the matrix marketplace, automatically promote
    integers to floats if we recognize them in the source.

    The specific functions we look for are:

        trigonometric: sin, asin, sinh, asinh, etc., and atan2
        exponential:   exp, exp2, exp10, expm1, log, log2, log10, logp1
        power:         pow, pown, powr, sqrt, rsqrt, rootn
        special:       erf, erfc, tgamma
        float:         fabs, fmin, fmax

    Note that we don't convert the second argument of dual argument
    functions: atan2, fmax, fmin, pow, powr.  This could potentially
    be a problem for pow(x, 2), but that case seems to work without change.
    """
    out = TGMATH_INT_RE.sub(r'\g<0>.', source)
    return out


# Floating point regular expression
#
# Define parts:
#
#    E = [eE][+-]?\d+    : Exponent
#    P = [.]             : Decimal separator
#    N = [1-9]\d*        : Natural number, no leading zeros
#    Z = 0               : Zero
#    F = \d+             : Fractional number, maybe leading zeros
#    F? = \d*            : Optional fractional number
#
# We want to reject bare natural numbers and bare decimal points, so we
# need to tediously outline the cases where we have either a fraction or
# an exponent:
#
#   ( ZP | ZPF | ZE | ZPE | ZPFE | NP | NPF | NE | NPE | NPFE | PF | PFE )
#
#
# We can then join cases by making parts optional.  The following are
# some ways to do this:
#
#   ( (Z|N)(P|PF|E|PE|PFE) | PFE? )                   # Split on lead
#     => ( (Z|N)(PF?|(PF?)?E) | PFE? )
#   ( ((Z|N)PF?|PF)E? | (Z|N)E)                       # Split on point
#   ( (ZP|ZPF|NP|NPF|PF) | (Z|ZP|ZPF|N|NP|NPF|PF)E )  # Split on E
#     => ( ((Z|N)PF?|PF) | ((Z|N)(PF?)? | PF) E )
FLOAT_RE = re.compile(r"""
    (?<!\w)  # use negative lookbehind since '.' confuses \b test
    # use split on lead to match float ( (Z|N)(PF?|(PF?)?E) | PFE? )
    ( ( 0 | [1-9]\d* )                     # ( ( Z | N )
      ([.]\d* | ([.]\d*)? [eE][+-]?\d+ )   #   (PF? | (PF?)? E )
    | [.]\d+ ([eE][+-]?\d+)?               # | PF (E)?
    )                                      # )
    (?!\w)  # use negative lookahead since '.' confuses \b test
    """, re.VERBOSE)
def _tag_float(source, constant_flag):
    # Convert floating point constants to single by adding 'f' to the end,
    # or long double with an 'L' suffix.  OS/X complains if you don't do this.
    out = FLOAT_RE.sub(r'\g<0>%s'%constant_flag, source)
    #print("in",repr(source),"out",repr(out), constant_flag)
    return out

def test_tag_float():
    """check that floating point constants are properly identified and tagged with 'f'"""

    cases = """
ZP  : 0.
ZPF : 0.0,0.01,0.1
Z  E: 0e+001
ZP E: 0.E0
ZPFE: 0.13e-031
NP  : 1., 12.
NPF : 1.0001, 1.1, 1.0
N  E: 1e0, 37E-080
NP E: 1.e0, 37.E-080
NPFE: 845.017e+22
 PF : .1, .0, .0100
 PFE: .6e+9, .82E-004
# isolated cases
0.
1e0
0.13e-013
# untouched
struct3.e3, 03.05.67, 37
# expressions
3.75+-1.6e-7-27+13.2
a3.e2 - 0.
4*atan(1)
4.*atan(1.)
"""

    output = """
ZP  : 0.f
ZPF : 0.0f,0.01f,0.1f
Z  E: 0e+001f
ZP E: 0.E0f
ZPFE: 0.13e-031f
NP  : 1.f, 12.f
NPF : 1.0001f, 1.1f, 1.0f
N  E: 1e0f, 37E-080f
NP E: 1.e0f, 37.E-080f
NPFE: 845.017e+22f
 PF : .1f, .0f, .0100f
 PFE: .6e+9f, .82E-004f
# isolated cases
0.f
1e0f
0.13e-013f
# untouched
struct3.e3, 03.05.67, 37
# expressions
3.75f+-1.6e-7f-27+13.2f
a3.e2 - 0.f
4*atan(1)
4.f*atan(1.f)
"""

    for case_in, case_out in zip(cases.split('\n'), output.split('\n')):
        out = _tag_float(case_in, 'f')
        assert case_out == out, "%r => %r"%(case_in, out)


def kernel_name(model_info, variant):
    # type: (ModelInfo, str) -> str
    """
    Name of the exported kernel symbol.

    *variant* is "Iq", "Iqxy" or "Imagnetic".
    """
    return model_info.name + "_" + variant


def indent(s, depth):
    # type: (str, int) -> str
    """
    Indent a string of text with *depth* additional spaces on each line.
    """
    spaces = " "*depth
    sep = "\n" + spaces
    return spaces + sep.join(s.split("\n"))


_template_cache = {}  # type: Dict[str, Tuple[int, str, str]]
def load_template(filename):
    # type: (str) -> str
    path = joinpath(DATA_PATH, filename)
    mtime = getmtime(path)
    if filename not in _template_cache or mtime > _template_cache[filename][0]:
        with open(path) as fid:
            _template_cache[filename] = (mtime, fid.read(), path)
    return _template_cache[filename][1], path


_FN_TEMPLATE = """\
double %(name)s(%(pars)s);
double %(name)s(%(pars)s) {
#line %(line)d "%(filename)s"
    %(body)s
}

"""
def _gen_fn(name, pars, body, filename, line):
    # type: (str, List[Parameter], str, str, int) -> str
    """
    Generate a function given pars and body.

    Returns the following string::

         double fn(double a, double b, ...);
         double fn(double a, double b, ...) {
             ....
         }
    """
    par_decl = ', '.join(p.as_function_argument() for p in pars) if pars else 'void'
    return _FN_TEMPLATE % {
        'name': name, 'pars': par_decl, 'body': body,
        'filename': filename.replace('\\', '\\\\'), 'line': line,
    }


def _call_pars(prefix, pars):
    # type: (str, List[Parameter]) -> List[str]
    """
    Return a list of *prefix+parameter* from parameter items.

    *prefix* should be "v." if v is a struct.
    """
    return [p.as_call_reference(prefix) for p in pars]


# type in IQXY pattern could be single, float, double, long double, ...
_IQXY_PATTERN = re.compile("^((inline|static) )? *([a-z ]+ )? *Iqxy *([(]|$)",
                           flags=re.MULTILINE)
def _have_Iqxy(sources):
    # type: (List[str]) -> bool
    """
    Return true if any file defines Iqxy.

    Note this is not a C parser, and so can be easily confused by
    non-standard syntax.  Also, it will incorrectly identify the following
    as having Iqxy::

        /*
        double Iqxy(qx, qy, ...) { ... fill this in later ... }
        */

    If you want to comment out an Iqxy function, use // on the front of the
    line instead.
    """
    for path, code in sources:
        if _IQXY_PATTERN.search(code):
            return True
    else:
        return False


def _add_source(source, code, path):
    """
    Add a file to the list of source code chunks, tagged with path and line.
    """
    path = path.replace('\\', '\\\\')
    source.append('#line 1 "%s"' % path)
    source.append(code)

def make_source(model_info):
    # type: (ModelInfo) -> Dict[str, str]
    """
    Generate the OpenCL/ctypes kernel from the module info.

    Uses source files found in the given search path.  Returns None if this
    is a pure python model, with no C source components.
    """
    if callable(model_info.Iq):
        raise ValueError("can't compile python model")
        #return None

    # TODO: need something other than volume to indicate dispersion parameters
    # No volume normalization despite having a volume parameter.
    # Thickness is labelled a volume in order to trigger polydispersity.
    # May want a separate dispersion flag, or perhaps a separate category for
    # disperse, but not volume.  Volume parameters also use relative values
    # for the distribution rather than the absolute values used by angular
    # dispersion.  Need to be careful that necessary parameters are available
    # for computing volume even if we allow non-disperse volume parameters.

    partable = model_info.parameters

    # Load templates and user code
    kernel_header = load_template('kernel_header.c')
    dll_code = load_template('kernel_iq.c')
    ocl_code = load_template('kernel_iq.cl')
    #ocl_code = load_template('kernel_iq_local.cl')
    user_code = [(f, open(f).read()) for f in model_sources(model_info)]

    # Build initial sources
    source = []
    _add_source(source, *kernel_header)
    for path, code in user_code:
        _add_source(source, code, path)

    # Make parameters for q, qx, qy so that we can use them in declarations
    q, qx, qy = [Parameter(name=v) for v in ('q', 'qx', 'qy')]
    # Generate form_volume function, etc. from body only
    if isinstance(model_info.form_volume, str):
        pars = partable.form_volume_parameters
        source.append(_gen_fn('form_volume', pars, model_info.form_volume,
                              model_info.filename, model_info._form_volume_line))
    if isinstance(model_info.Iq, str):
        pars = [q] + partable.iq_parameters
        source.append(_gen_fn('Iq', pars, model_info.Iq,
                              model_info.filename, model_info._Iq_line))
    if isinstance(model_info.Iqxy, str):
        pars = [qx, qy] + partable.iqxy_parameters
        source.append(_gen_fn('Iqxy', pars, model_info.Iqxy,
                              model_info.filename, model_info._Iqxy_line))

    # Define the parameter table
    # TODO: plug in current line number
    source.append('#line 542 "sasmodels/generate.py"')
    source.append("#define PARAMETER_TABLE \\")
    source.append("\\\n".join(p.as_definition()
                              for p in partable.kernel_parameters))

    # Define the function calls
    if partable.form_volume_parameters:
        refs = _call_pars("_v.", partable.form_volume_parameters)
        call_volume = "#define CALL_VOLUME(_v) form_volume(%s)"%(",".join(refs))
    else:
        # Model doesn't have volume.  We could make the kernel run a little
        # faster by not using/transferring the volume normalizations, but
        # the ifdef's reduce readability more than is worthwhile.
        call_volume = "#define CALL_VOLUME(v) 1.0"
    source.append(call_volume)

    refs = ["_q[_i]"] + _call_pars("_v.", partable.iq_parameters)
    call_iq = "#define CALL_IQ(_q,_i,_v) Iq(%s)" % (",".join(refs))
    if _have_Iqxy(user_code) or isinstance(model_info.Iqxy, str):
        # Call 2D model
        refs = ["_q[2*_i]", "_q[2*_i+1]"] + _call_pars("_v.", partable.iqxy_parameters)
        call_iqxy = "#define CALL_IQ(_q,_i,_v) Iqxy(%s)" % (",".join(refs))
    else:
        # Call 1D model with sqrt(qx^2 + qy^2)
        #warnings.warn("Creating Iqxy = Iq(sqrt(qx^2 + qy^2))")
        # still defined:: refs = ["q[i]"] + _call_pars("v", iq_parameters)
        pars_sqrt = ["sqrt(_q[2*_i]*_q[2*_i]+_q[2*_i+1]*_q[2*_i+1])"] + refs[1:]
        call_iqxy = "#define CALL_IQ(_q,_i,_v) Iq(%s)" % (",".join(pars_sqrt))

    magpars = [k-2 for k, p in enumerate(partable.call_parameters)
               if p.type == 'sld']

    # Fill in definitions for numbers of parameters
    source.append("#define MAX_PD %s"%partable.max_pd)
    source.append("#define NUM_PARS %d"%partable.npars)
    source.append("#define NUM_VALUES %d" % partable.nvalues)
    source.append("#define NUM_MAGNETIC %d" % partable.nmagnetic)
    source.append("#define MAGNETIC_PARS %s"%",".join(str(k) for k in magpars))
    for k, v in enumerate(magpars[:3]):
        source.append("#define MAGNETIC_PAR%d %d"%(k+1, v))

    # TODO: allow mixed python/opencl kernels?

    ocl = kernels(ocl_code, call_iq, call_iqxy, model_info.name)
    dll = kernels(dll_code, call_iq, call_iqxy, model_info.name)
    result = {
        'dll': '\n'.join(source+dll[0]+dll[1]+dll[2]),
        'opencl': '\n'.join(source+ocl[0]+ocl[1]+ocl[2]),
    }

    return result


def kernels(kernel, call_iq, call_iqxy, name):
    # type: ([str,str], str, str, str) -> List[str]
    code = kernel[0]
    path = kernel[1].replace('\\', '\\\\')
    iq = [
        # define the Iq kernel
        "#define KERNEL_NAME %s_Iq" % name,
        call_iq,
        '#line 1 "%s Iq"' % path,
        code,
        "#undef CALL_IQ",
        "#undef KERNEL_NAME",
        ]

    iqxy = [
        # define the Iqxy kernel from the same source with different #defines
        "#define KERNEL_NAME %s_Iqxy" % name,
        call_iqxy,
        '#line 1 "%s Iqxy"' % path,
        code,
        "#undef CALL_IQ",
        "#undef KERNEL_NAME",
    ]

    imagnetic = [
        # define the Imagnetic kernel
        "#define KERNEL_NAME %s_Imagnetic" % name,
        "#define MAGNETIC 1",
        call_iqxy,
        '#line 1 "%s Imagnetic"' % path,
        code,
        "#undef MAGNETIC",
        "#undef CALL_IQ",
        "#undef KERNEL_NAME",
    ]

    return iq, iqxy, imagnetic


def load_kernel_module(model_name):
    # type: (str) -> module
    """
    Return the kernel module named in *model_name*.

    If the name ends in *.py* then load it as a custom model using
    :func:`sasmodels.custom.load_custom_kernel_module`, otherwise
    load it from :mod:`sasmodels.models`.
    """
    if model_name.endswith('.py'):
        kernel_module = load_custom_kernel_module(model_name)
    else:
        from sasmodels import models
        __import__('sasmodels.models.'+model_name)
        kernel_module = getattr(models, model_name, None)
    return kernel_module


section_marker = re.compile(r'\A(?P<first>[%s])(?P=first)*\Z'
                            % re.escape(string.punctuation))
def _convert_section_titles_to_boldface(lines):
    # type: (Sequence[str]) -> Iterator[str]
    """
    Do the actual work of identifying and converting section headings.
    """
    prior = None
    for line in lines:
        if prior is None:
            prior = line
        elif section_marker.match(line):
            if len(line) >= len(prior):
                yield "".join(("**", prior, "**"))
                prior = None
            else:
                yield prior
                prior = line
        else:
            yield prior
            prior = line
    if prior is not None:
        yield prior


def convert_section_titles_to_boldface(s):
    # type: (str) -> str
    """
    Use explicit bold-face rather than section headings so that the table of
    contents is not polluted with section names from the model documentation.

    Sections are identified as the title line followed by a line of punctuation
    at least as long as the title line.
    """
    return "\n".join(_convert_section_titles_to_boldface(s.split('\n')))


def make_doc(model_info):
    # type: (ModelInfo) -> str
    """
    Return the documentation for the model.
    """
    Iq_units = "The returned value is scaled to units of |cm^-1| |sr^-1|, absolute scale."
    Sq_units = "The returned value is a dimensionless structure factor, $S(q)$."
    docs = model_info.docs if model_info.docs is not None else ""
    docs = convert_section_titles_to_boldface(docs)
    pars = make_partable(model_info.parameters.COMMON
                         + model_info.parameters.kernel_parameters)
    subst = dict(id=model_info.id.replace('_', '-'),
                 name=model_info.name,
                 title=model_info.title,
                 parameters=pars,
                 returns=Sq_units if model_info.structure_factor else Iq_units,
                 docs=docs)
    return DOC_HEADER % subst


# TODO: need a single source for rst_prolog; it is also in doc/rst_prolog
RST_PROLOG = r"""\
.. |Ang| unicode:: U+212B
.. |Ang^-1| replace:: |Ang|\ :sup:`-1`
.. |Ang^2| replace:: |Ang|\ :sup:`2`
.. |Ang^-2| replace:: |Ang|\ :sup:`-2`
.. |1e-6Ang^-2| replace:: 10\ :sup:`-6`\ |Ang|\ :sup:`-2`
.. |Ang^3| replace:: |Ang|\ :sup:`3`
.. |Ang^-3| replace:: |Ang|\ :sup:`-3`
.. |Ang^-4| replace:: |Ang|\ :sup:`-4`
.. |cm^-1| replace:: cm\ :sup:`-1`
.. |cm^2| replace:: cm\ :sup:`2`
.. |cm^-2| replace:: cm\ :sup:`-2`
.. |cm^3| replace:: cm\ :sup:`3`
.. |1e15cm^3| replace:: 10\ :sup:`15`\ cm\ :sup:`3`
.. |cm^-3| replace:: cm\ :sup:`-3`
.. |sr^-1| replace:: sr\ :sup:`-1`

.. |cdot| unicode:: U+00B7
.. |deg| unicode:: U+00B0
.. |g/cm^3| replace:: g\ |cdot|\ cm\ :sup:`-3`
.. |mg/m^2| replace:: mg\ |cdot|\ m\ :sup:`-2`
.. |fm^2| replace:: fm\ :sup:`2`
.. |Ang*cm^-1| replace:: |Ang|\ |cdot|\ cm\ :sup:`-1`
"""

# TODO: make a better fake reference role
RST_ROLES = """\
.. role:: ref

.. role:: numref

"""

def make_html(model_info):
    """
    Convert model docs directly to html.
    """
    from . import rst2html

    rst = make_doc(model_info)
    return rst2html.rst2html("".join((RST_ROLES, RST_PROLOG, rst)))

def view_html(model_name):
    from . import modelinfo
    kernel_module = load_kernel_module(model_name)
    info = modelinfo.make_model_info(kernel_module)
    view_html_from_info(info)

def view_html_from_info(info):
    from . import rst2html
    url = "file://"+dirname(info.filename)+"/"
    rst2html.view_html(make_html(info), url=url)

def demo_time():
    # type: () -> None
    """
    Show how long it takes to process a model.
    """
    import datetime
    from .modelinfo import make_model_info
    from .models import cylinder

    tic = datetime.datetime.now()
    make_source(make_model_info(cylinder))
    toc = (datetime.datetime.now() - tic).total_seconds()
    print("time: %g"%toc)


def main():
    # type: () -> None
    """
    Program which prints the source produced by the model.
    """
    import sys
    from .modelinfo import make_model_info

    if len(sys.argv) <= 1:
        print("usage: python -m sasmodels.generate modelname")
    else:
        name = sys.argv[1]
        kernel_module = load_kernel_module(name)
        model_info = make_model_info(kernel_module)
        source = make_source(model_info)
        print(source['dll'])


if __name__ == "__main__":
    main()



r"""
Calculates the scattering from a randomly distributed, two-phase system based on
the Debye-Anderson-Brumberger (DAB) model for such systems. The two-phase system
is characterized by a single length scale, the correlation length, which is a
measure of the average spacing between regions of phase 1 and phase 2. **The
model also assumes smooth interfaces between the phases** and hence exhibits
Porod behavior $(I \sim q^{-4})$ at large $q$, $(qL \gg 1)$.

The DAB model is ostensibly a development of the earlier Debye-Bueche model.

Definition
----------

.. math::

    I(q) = \text{scale}\cdot\frac{L^3}{(1 + (q\cdot L)^2)^2} + \text{background}

where scale is

.. math:: \text{scale} = 8 \pi \phi (1-\phi) \Delta\rho^2

and the parameter $L$ is the correlation length.

For 2D data the scattering intensity is calculated in the same way as 1D,
where the $q$ vector is defined as

.. math:: q = \sqrt{q_x^2 + q_y^2}


References
----------

P Debye, H R Anderson, H Brumberger, *Scattering by an Inhomogeneous Solid. II.
The Correlation Function and its Application*, *J. Appl. Phys.*, 28(6) (1957) 679

P Debye, A M Bueche, *Scattering by an Inhomogeneous Solid*, *J. Appl. Phys.*,
20 (1949) 518

*2013/09/09 - Description reviewed by King, S and Parker, P.*

"""

from numpy import inf

name = "dab"
title = "DAB (Debye Anderson Brumberger) Model"
description = """\

F(q)= scale * L^3/(1 + (q*L)^2)^2

L: the correlation length

"""
category = "shape-independent"

#             ["name", "units", default, [lower, upper], "type", "description"],
parameters = [["cor_length", "Ang", 50.0, [0, inf], "", "correlation length"],
             ]

Iq = """
    double numerator   = cube(cor_length);
    double denominator = square(1 + square(q*cor_length));

    return numerator / denominator ;
    """

def random():
    import numpy as np
    pars = dict(
        scale=10**np.random.uniform(1, 4),
        cor_length=10**np.random.uniform(0.3, 3),
        #background = 0,
    )
    pars['scale'] /= pars['cor_length']**3
    return pars

# ER defaults to 1.0

# VR defaults to 1.0

demo = dict(scale=1, background=0, cor_length=50)



"""
Minimal example of calling a kernel for a specific set of q values.
"""

from numpy import logspace
from matplotlib import pyplot as plt
from sasmodels.core import load_model
from sasmodels.direct_model import call_kernel

model = load_model('cylinder')
q = logspace(-3, -1, 200)
kernel = model.make_kernel([q])
Iq = call_kernel(kernel, dict(radius=200.))
plt.loglog(q, Iq)
plt.xlabel('q (1/A)')
plt.ylabel('I(q)')
plt.title('Cylinder with radius 200.')
plt.show()



#!/usr/bin/env python

# Copyright (C) 2006-2010, University of Maryland
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/ or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Author: James Krycka

"""
This script uses py2exe to create dist\bumps.exe and dist\bumps_gui.exe for
running the Bumps application in either CLI or GUI mode.

These executables start the application and import the rest of the application
code stored in library.zip.  The python interpreter and other required python
packages and dlls are also placed in the zip file.  Additional resource files
that are needed when Bumps is run are copied to the dist directory tree.  On
completion, the contents of the dist directory tree can be used by the Inno
Setup Compiler (via a separate script) to build a Windows installer/uninstaller
for deployment of the Bumps application.  For testing purposes, bumps.exe or
bumps_gui.exe can be run from the dist directory.
"""

import os
import sys

#sys.dont_write_bytecode = True

# Force build before continuing
os.system('"%s" setup.py build' % sys.executable)

# Remove the current directory from the python path
here = os.path.abspath(os.path.dirname(__file__))
sys.path = [p for p in sys.path if os.path.abspath(p) != here]

import glob

from distutils.core import setup
from distutils.util import get_platform

# Augment the setup interface with the py2exe command and make sure the py2exe
# option is passed to setup.
import py2exe

if len(sys.argv) == 1:
    sys.argv.append('py2exe')

# Put the build lib on the start of the path.
# For packages with binary extensions, need platform.  If it is a pure
# script library, use an empty platform string.
platform = '.%s-%s' % (get_platform(), sys.version[:3])
#platform = ''
build_lib = os.path.abspath('build/lib' + platform)
sys.path.insert(0, build_lib)

# print "\n".join(sys.path)

import wx
import matplotlib
matplotlib.use('WXAgg')
import periodictable

# Retrieve the application version string.
import bumps
version = bumps.__version__
from bumps.gui.resources import resources as gui_resources

# A manifest is required to be included in a py2exe image (or accessible as a
# file in the image directory) when wxPython is included so that the Windows XP
# theme is used when rendering wx widgets.  The manifest must be matched to the
# version of Python that is being used.
#
# Create a manifest for use with Python 2.5 on Windows XP or Vista.  It is
# adapted from the Python manifest file (C:\Python25\pythonw.exe.manifest).

manifest_for_python25 = """
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<assembly xmlns="urn:schemas-microsoft-com:asm.v1" manifestVersion="1.0">
<assemblyIdentity
    version="1.0.0.0"
    processorArchitecture="x86"
    name="%(prog)s"
    type="win32"
/>
<description>%(prog)s</description>
<dependency>
    <dependentAssembly>
        <assemblyIdentity
            type="win32"
            name="Microsoft.Windows.Common-Controls"
            version="6.0.0.0"
            processorArchitecture="X86"
            publicKeyToken="6595b64144ccf1df"
            language="*"
        />
    </dependentAssembly>
</dependency>
</assembly>
"""

# Create a manifest for use with Python 2.6 or 2.7 on Windows XP or Vista.

manifest_for_python26 = """
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<assembly xmlns="urn:schemas-microsoft-com:asm.v1" manifestVersion="1.0">
  <assemblyIdentity
    version="5.0.0.0"
    processorArchitecture="x86"
    name="%(prog)s"
    type="win32">
  </assemblyIdentity>
  <description>%(prog)s</description>
  <trustInfo xmlns="urn:schemas-microsoft-com:asm.v3">
    <security>
      <requestedPrivileges>
        <requestedExecutionLevel
          level="asInvoker"
          uiAccess="false">
        </requestedExecutionLevel>
      </requestedPrivileges>
    </security>
  </trustInfo>
  <dependency>
    <dependentAssembly>
      <assemblyIdentity
        type="win32"
        name="Microsoft.VC90.CRT"
        version="9.0.21022.8"
        processorArchitecture="x86"
        publicKeyToken="1fc8b3b9a1e18e3b">
      </assemblyIdentity>
    </dependentAssembly>
  </dependency>
  <dependency>
    <dependentAssembly>
      <assemblyIdentity
        type="win32"
        name="Microsoft.Windows.Common-Controls"
        version="6.0.0.0"
        processorArchitecture="x86"
        publicKeyToken="6595b64144ccf1df"
        language="*">
      </assemblyIdentity>
    </dependentAssembly>
  </dependency>
</assembly>
"""

# Select the appropriate manifest to use.
if sys.version_info >= (3, 0) or sys.version_info < (2, 5):
    print("*** This script only works with Python 2.5, 2.6, or 2.7.")
    sys.exit()
elif sys.version_info >= (2, 6):
    manifest = manifest_for_python26
elif sys.version_info >= (2, 5):
    manifest = manifest_for_python25

# Create a list of all files to include along side the executable being built
# in the dist directory tree.  Each element of the data_files list is a tuple
# consisting of a path (relative to dist\) and a list of files in that path.
data_files = []

# Add resource files that need to reside in the same directory as the image.
data_files.append(('.', [os.path.join('.', 'LICENSE.txt')]))
data_files.append(('.', [os.path.join('.', 'README.txt')]))
data_files.append(('.', [os.path.join('.', 'bin', 'bumps_launch.bat')]))

# Add application specific data files from the bumps\bumps-data folder.
data_files += gui_resources.data_files()

# Add data files from the matplotlib\mpl-data folder and its subfolders.
# For matploblib prior to version 0.99 see the examples at the end of the file.
data_files += matplotlib.get_py2exe_datafiles()

# Add data files from the periodictable\xsf folder.
data_files += periodictable.data_files()

# Add example directories and their files.  An empty directory is ignored.
# Note that Inno Setup will determine where these files will be placed such as
# C:\My Documents\... instead of the installation folder.
for path in glob.glob(os.path.join('examples', '*')):
    if os.path.isdir(path):
        for file in glob.glob(os.path.join(path, '*.*')):
            data_files.append((path, [file]))
    else:
        data_files.append(('examples', [path]))

for path in glob.glob(os.path.join('doc', 'examples', '*')):
    if os.path.isdir(path):
        for file in glob.glob(os.path.join(path, '*.*')):
            data_files.append((path, [file]))
    else:
        data_files.append(('doc', [path]))

# Add PDF documentation to the dist staging directory.
pdf = os.path.join('doc', 'Bumps.pdf')
if os.path.isfile(pdf):
    data_files.append(('doc', [pdf]))
else:
    print("*** %s not found - building frozen image without it ***" % pdf)

# Add the Microsoft Visual C++ 2008 redistributable kit if we are building with
# Python 2.6 or 2.7.  This kit will be installed on the target system as part
# of the installation process for the frozen image.  Note that the Python 2.5
# interpreter requires msvcr71.dll which is included in the Python25 package,
# however, Python 2.6 and 2.7 require the msvcr90.dll but they do not bundle it
# with the Python26 or Python27 package.  Thus, for Python 2.6 and later, the
# appropriate dll must be present on the target system at runtime.
if sys.version_info >= (2, 6):
    pypath = os.path.dirname(sys.executable)
    data_files.append(('.', [os.path.join(pypath, 'vcredist_x86.exe')]))

# Specify required packages to bundle in the executable image.
packages = ['numpy', 'scipy', 'matplotlib', 'pytz', 'pyparsing',
            'periodictable', 'bumps', 'sasmodels', 'pyopencl',
            ]

# Specify files to include in the executable image.
includes = []

# Specify files to exclude from the executable image.
# - We can safely exclude Tk/Tcl and Qt modules because our app uses wxPython.
# - We do not use ssl services so they are omitted.
# - We can safely exclude the TkAgg matplotlib backend because our app uses
#   "matplotlib.use('WXAgg')" to override the default matplotlib configuration.
# - On the web it is widely recommended to exclude certain lib*.dll modules
#   but this does not seem necessary any more (but adding them does not hurt).
# - Python25 requires mscvr71.dll, however, Win XP includes this file.
# - Since we do not support Win 9x systems, w9xpopen.dll is not needed.
# - For some reason cygwin1.dll gets included by default, but it is not needed.

excludes = ['Tkinter', 'PyQt4', '_ssl', '_tkagg', 'numpy.distutils.test']

dll_excludes = ['libgdk_pixbuf-2.0-0.dll',
                'libgobject-2.0-0.dll',
                'libgdk-win32-2.0-0.dll',
                'tcl84.dll',
                'tk84.dll',
                'QtGui4.dll',
                'QtCore4.dll',
                'msvcr71.dll',
                'msvcp90.dll',
                'w9xpopen.exe',
                'cygwin1.dll']


class Target(object):

    """This class stores metadata about the distribution in a dictionary."""

    def __init__(self, **kw):
        self.__dict__.update(kw)
        self.version = version

clientCLI = Target(
    name='Bumps',
    description='Bumps CLI application',
    # module to run on application start
    script=os.path.join('bin', 'bumps_cli.py'),
    dest_base='bumps',  # file name part of the exe file to create
    # also need to specify in data_files
    icon_resources=[
        (1, os.path.join('bumps', 'gui', 'resources', 'bumps.ico'))],
    bitmap_resources=[],
    other_resources=[(24, 1, manifest % dict(prog='Bumps'))])

clientGUI = Target(
    name='Bumps',
    description='Bumps GUI application',
    # module to run on application start
    script=os.path.join('bin', 'bumps_gui.py'),
    dest_base='bumps_gui',  # file name part of the exe file to create
    # also need to specify in data_files
    icon_resources=[
        (1, os.path.join('bumps', 'gui', 'resources', 'bumps.ico'))],
    bitmap_resources=[],
    other_resources=[(24, 1, manifest % dict(prog='Bumps'))])

# Now we do the work to create a standalone distribution using py2exe.
#
# When the application is run in console mode, a console window will be created
# to receive any logging or error messages and the application will then create
# a separate GUI application window.
#
# When the application is run in windows mode, it will create a GUI application
# window and no console window will be provided.  Output to stderr will be
# written to <app-image-name>.log.
setup(
    console=[clientCLI],
    windows=[clientGUI],
    options={'py2exe': {
        'packages': packages,
        'includes': includes,
        'excludes': excludes,
        'dll_excludes': dll_excludes,
        'compressed': 1,   # standard compression
        'optimize': 0,     # no byte-code optimization
        'dist_dir': "dist",  # where to put py2exe results
        'xref': False,     # display cross reference (as html doc)
        'bundle_files': 1,  # bundle python25.dll in library.zip
    }
    },
    # Since we are building two exe's, do not put the shared library in each
    # of them.  Instead create a single, separate library.zip file.
    # zipfile=None,               # bundle library.zip in exe
    data_files=data_files           # list of files to copy to dist directory
)



"""
Utility to add annotations to python exceptions.
"""
import sys

# Platform cruft: WindowsError is only defined on Windows.
try:
    WindowsError
except NameError:
    class WindowsError(Exception):
        """
        Fake WindowsException when not on Windows.
        """
        pass

def annotate_exception(msg, exc=None):
    """
    Add an annotation to the current exception, which can then be forwarded
    to the caller using a bare "raise" statement to raise the annotated
    exception.  If the exception *exc* is provided, then that exception is the
    one that is annotated, otherwise *sys.exc_info* is used.

    Example::

        >>> D = {}
        >>> try:
        ...    print(D['hello'])
        ... except:
        ...    annotate_exception("while accessing 'D'")
        ...    raise
        Traceback (most recent call last):
            ...
        KeyError: "hello while accessing 'D'"
    """
    if not exc:
        exc = sys.exc_info()[1]

    # Can't extend WindowsError exceptions; instead raise a new exception.
    # TODO: try to incorporate current stack trace in the raised exception
    if isinstance(exc, WindowsError):
        raise OSError(str(exc) + " " + msg)

    args = exc.args
    if not args:
        exc.args = (msg,)
    else:
        try:
            arg0 = " ".join((args[0], msg))
            exc.args = tuple([arg0] + list(args[1:]))
        except:
            exc.args = (" ".join((str(exc), msg)),)



from __future__ import print_function

import sys
# make sure sasmodels is on the path
sys.path.append('..')

from os import mkdir
from os.path import basename, exists, join as joinpath
from sasmodels.core import load_model_info

try:
    from typing import Optional, BinaryIO, List, Dict
except ImportError:
    pass
else:
    from sasmodels.modelinfo import ModelInfo

TEMPLATE = """\
..
    Generated from doc/gentoc.py -- DO NOT EDIT --

.. _%(label)s:

%(bar)s
%(title)s
%(bar)s

.. toctree::

"""

MODEL_TOC_PATH = "guide/models"

def _make_category(category_name, label, title, parent=None):
    # type: (str, str, str, Optional[BinaryIO]) -> BinaryIO
    file = open(joinpath(MODEL_TOC_PATH, category_name+".rst"), "w")
    file.write(TEMPLATE%{'label':label, 'title':title, 'bar':'*'*len(title)})
    if parent:
        _add_subcategory(category_name, parent)
    return file

def _add_subcategory(category_name, parent):
    # type: (str, BinaryIO) -> None
    parent.write("    %s.rst\n"%category_name)

def _add_model(file, model_name):
    # type: (IO[str], str) -> None
    file.write("    ../../model/%s.rst\n"%model_name)

def _maybe_make_category(category, models, cat_files, model_toc):
    # type: (str, List[str], Dict[str, BinaryIO], BinaryIO) -> None
    if category not in cat_files:
        print("Unexpected category %s containing"%category, models, file=sys.stderr)
        title = category.capitalize()+" Functions"
        cat_files[category] = _make_category(category, category, title, model_toc)

def generate_toc(model_files):
    # type: (List[str]) -> None
    if not model_files:
        print("gentoc needs a list of model files", file=sys.stderr)

    # find all categories
    category = {} # type: Dict[str, List[str]]
    for item in model_files:
        # assume model is in sasmodels/models/name.py, and ignore the full path
        model_name = basename(item)[:-3]
        if model_name.startswith('_'):
            continue
        model_info = load_model_info(model_name)
        if model_info.category is None:
            print("Missing category for", item, file=sys.stderr)
        else:
            category.setdefault(model_info.category, []).append(model_name)

    # Check category names
    for k, v in category.items():
        if len(v) == 1:
            print("Category %s contains only %s"%(k, v[0]), file=sys.stderr)

    # Generate category files for the table of contents.
    # Initially we had "shape functions" as an additional TOC level, but we
    # have revised it so that the individual shape categories now go at
    # the top level.  Judicious rearrangement of comments will make the
    # "shape functions" level reappear.
    # We are forcing shape-independent, structure-factor and custom-models
    # to come at the end of the TOC.  All other categories will come in
    # alphabetical order before them.

    if not exists(MODEL_TOC_PATH):
        mkdir(MODEL_TOC_PATH)
    model_toc = _make_category(
        'index', 'Models', 'Model Functions')
    #shape_toc = _make_category(
    #    'shape',  'Shapes', 'Shape Functions', model_toc)
    free_toc = _make_category(
        'shape-independent', 'Shape-independent',
        'Shape-Independent Functions')
    struct_toc = _make_category(
        'structure-factor', 'Structure-factor', 'Structure Factors')
    #custom_toc = _make_category(
    #    'custom-models', 'Custom-models', 'Custom Models')

    # remember to top level categories
    cat_files = {
        #'shape':shape_toc,
        'shape':model_toc,
        'shape-independent':free_toc,
        'structure-factor': struct_toc,
        #'custom': custom_toc,
        }

    # Process the model lists
    for k, v in sorted(category.items()):
        if ':' in k:
            cat, subcat = k.split(':')
            _maybe_make_category(cat, v, cat_files, model_toc)
            cat_file = cat_files[cat]
            label = "-".join((cat, subcat))
            filename = label
            title = subcat.capitalize() + " Functions"
            sub_toc = _make_category(filename, label, title, cat_file)
            for model in sorted(v):
                _add_model(sub_toc, model)
            sub_toc.close()
        else:
            _maybe_make_category(k, v, cat_files, model_toc)
            cat_file = cat_files[k]
            for model in sorted(v):
                _add_model(cat_file, model)

    #_add_subcategory('shape', model_toc)
    _add_subcategory('shape-independent', model_toc)
    _add_subcategory('structure-factor', model_toc)
    #_add_subcategory('custom-models', model_toc)

    # Close the top-level category files
    #model_toc.close()
    for f in cat_files.values():
        f.close()


if __name__ == "__main__":
    generate_toc(sys.argv[1:])



r"""
Definition
----------

Calculates the form factor for a rectangular solid with a core-shell structure.
The thickness and the scattering length density of the shell or
"rim" can be different on each (pair) of faces. However at this time
the 1D calculation does **NOT** actually calculate a c face rim despite the presence of
the parameter. Some other aspects of the 1D calculation may be wrong.

.. note::
   This model was originally ported from NIST IGOR macros. However, it is not
   yet fully understood by the SasView developers and is currently under review.

The form factor is normalized by the particle volume $V$ such that

.. math::

    I(q) = \text{scale}\frac{\langle f^2 \rangle}{V} + \text{background}

where $\langle \ldots \rangle$ is an average over all possible orientations
of the rectangular solid.


The function calculated is the form factor of the rectangular solid below.
The core of the solid is defined by the dimensions $A$, $B$, $C$ such that
$A < B < C$.

.. image:: img/core_shell_parallelepiped_geometry.jpg

There are rectangular "slabs" of thickness $t_A$ that add to the $A$ dimension
(on the $BC$ faces). There are similar slabs on the $AC$ $(=t_B)$ and $AB$
$(=t_C)$ faces. The projection in the $AB$ plane is then

.. image:: img/core_shell_parallelepiped_projection.jpg

The volume of the solid is

.. math::

    V = ABC + 2t_ABC + 2t_BAC + 2t_CAB

**meaning that there are "gaps" at the corners of the solid.**  Again note that
$t_C = 0$ currently.

The intensity calculated follows the :ref:`parallelepiped` model, with the
core-shell intensity being calculated as the square of the sum of the
amplitudes of the core and shell, in the same manner as a core-shell model.

.. math::

    F_{a}(Q,\alpha,\beta)=
    \left[\frac{\sin(\tfrac{1}{2}Q(L_A+2t_A)\sin\alpha \sin\beta)}{\tfrac{1}{2}Q(L_A+2t_A)\sin\alpha\sin\beta}
    - \frac{\sin(\tfrac{1}{2}QL_A\sin\alpha \sin\beta)}{\tfrac{1}{2}QL_A\sin\alpha \sin\beta} \right]
    \left[\frac{\sin(\tfrac{1}{2}QL_B\sin\alpha \sin\beta)}{\tfrac{1}{2}QL_B\sin\alpha \sin\beta} \right]
    \left[\frac{\sin(\tfrac{1}{2}QL_C\sin\alpha \sin\beta)}{\tfrac{1}{2}QL_C\sin\alpha \sin\beta} \right]

.. note::

    Why does t_B not appear in the above equation?
    For the calculation of the form factor to be valid, the sides of the solid
    MUST (perhaps not any more?) be chosen such that** $A < B < C$.
    If this inequality is not satisfied, the model will not report an error,
    but the calculation will not be correct and thus the result wrong.

FITTING NOTES
If the scale is set equal to the particle volume fraction, $\phi$, the returned
value is the scattered intensity per unit volume, $I(q) = \phi P(q)$.
However, **no interparticle interference effects are included in this
calculation.**

There are many parameters in this model. Hold as many fixed as possible with
known values, or you will certainly end up at a solution that is unphysical.

Constraints must be applied during fitting to ensure that the inequality
$A < B < C$ is not violated. The calculation will not report an error,
but the results will not be correct.

The returned value is in units of |cm^-1|, on absolute scale.

NB: The 2nd virial coefficient of the core_shell_parallelepiped is calculated
based on the the averaged effective radius $(=\sqrt{(A+2t_A)(B+2t_B)/\pi})$
and length $(C+2t_C)$ values, after appropriately
sorting the three dimensions to give an oblate or prolate particle, to give an
effective radius, for $S(Q)$ when $P(Q) * S(Q)$ is applied.

To provide easy access to the orientation of the parallelepiped, we define the
axis of the cylinder using three angles $\theta$, $\phi$ and $\Psi$.
(see :ref:`cylinder orientation <cylinder-angle-definition>`).
The angle $\Psi$ is the rotational angle around the *long_c* axis against the
$q$ plane. For example, $\Psi = 0$ when the *short_b* axis is parallel to the
*x*-axis of the detector.

.. figure:: img/parallelepiped_angle_definition.png

    Definition of the angles for oriented core-shell parallelepipeds.

.. figure:: img/parallelepiped_angle_projection.png

    Examples of the angles for oriented core-shell parallelepipeds against the
    detector plane.

References
----------

.. [#] P Mittelbach and G Porod, *Acta Physica Austriaca*, 14 (1961) 185-211
    Equations (1), (13-14). (in German)
.. [#] D Singh (2009). *Small angle scattering studies of self assembly in
   lipid mixtures*, John's Hopkins University Thesis (2009) 223-225. `Available
   from Proquest <http://search.proquest.com/docview/304915826?accountid
   =26379>`_

Authorship and Verification
----------------------------

* **Author:** NIST IGOR/DANSE **Date:** pre 2010
* **Converted to sasmodels by:** Miguel Gonzales **Date:** February 26, 2016
* **Last Modified by:** Wojciech Potrzebowski **Date:** January 11, 2017
* **Currently Under review by:** Paul Butler
"""

import numpy as np
from numpy import pi, inf, sqrt, cos, sin

name = "core_shell_parallelepiped"
title = "Rectangular solid with a core-shell structure."
description = """
     P(q)=
"""
category = "shape:parallelepiped"

#             ["name", "units", default, [lower, upper], "type","description"],
parameters = [["sld_core", "1e-6/Ang^2", 1, [-inf, inf], "sld",
               "Parallelepiped core scattering length density"],
              ["sld_a", "1e-6/Ang^2", 2, [-inf, inf], "sld",
               "Parallelepiped A rim scattering length density"],
              ["sld_b", "1e-6/Ang^2", 4, [-inf, inf], "sld",
               "Parallelepiped B rim scattering length density"],
              ["sld_c", "1e-6/Ang^2", 2, [-inf, inf], "sld",
               "Parallelepiped C rim scattering length density"],
              ["sld_solvent", "1e-6/Ang^2", 6, [-inf, inf], "sld",
               "Solvent scattering length density"],
              ["length_a", "Ang", 35, [0, inf], "volume",
               "Shorter side of the parallelepiped"],
              ["length_b", "Ang", 75, [0, inf], "volume",
               "Second side of the parallelepiped"],
              ["length_c", "Ang", 400, [0, inf], "volume",
               "Larger side of the parallelepiped"],
              ["thick_rim_a", "Ang", 10, [0, inf], "volume",
               "Thickness of A rim"],
              ["thick_rim_b", "Ang", 10, [0, inf], "volume",
               "Thickness of B rim"],
              ["thick_rim_c", "Ang", 10, [0, inf], "volume",
               "Thickness of C rim"],
              ["theta", "degrees", 0, [-360, 360], "orientation",
               "c axis to beam angle"],
              ["phi", "degrees", 0, [-360, 360], "orientation",
               "rotation about beam"],
              ["psi", "degrees", 0, [-360, 360], "orientation",
               "rotation about c axis"],
             ]

source = ["lib/gauss76.c", "core_shell_parallelepiped.c"]


def ER(length_a, length_b, length_c, thick_rim_a, thick_rim_b, thick_rim_c):
    """
        Return equivalent radius (ER)
    """

    # surface average radius (rough approximation)
    surf_rad = sqrt((length_a + 2.0*thick_rim_a) * (length_b + 2.0*thick_rim_b) / pi)

    height = length_c + 2.0*thick_rim_c

    ddd = 0.75 * surf_rad * (2 * surf_rad * height + (height + surf_rad) * (height + pi * surf_rad))
    return 0.5 * (ddd) ** (1. / 3.)

# VR defaults to 1.0

def random():
    import numpy as np
    outer = 10**np.random.uniform(1, 4.7, size=3)
    thick = np.random.beta(0.5, 0.5, size=3)*(outer-2) + 1
    length = outer - thick
    pars = dict(
        length_a=length[0],
        length_b=length[1],
        length_c=length[2],
        thick_rim_a=thick[0],
        thick_rim_b=thick[1],
        thick_rim_c=thick[2],
    )
    return pars

# parameters for demo
demo = dict(scale=1, background=0.0,
            sld_core=1, sld_a=2, sld_b=4, sld_c=2, sld_solvent=6,
            length_a=35, length_b=75, length_c=400,
            thick_rim_a=10, thick_rim_b=10, thick_rim_c=10,
            theta=0, phi=0, psi=0,
            length_a_pd=0.1, length_a_pd_n=1,
            length_b_pd=0.1, length_b_pd_n=1,
            length_c_pd=0.1, length_c_pd_n=1,
            thick_rim_a_pd=0.1, thick_rim_a_pd_n=1,
            thick_rim_b_pd=0.1, thick_rim_b_pd_n=1,
            thick_rim_c_pd=0.1, thick_rim_c_pd_n=1,
            theta_pd=10, theta_pd_n=1,
            phi_pd=10, phi_pd_n=1,
            psi_pd=10, psi_pd_n=1)

# rkh 7/4/17 add random unit test for 2d, note make all params different, 2d values not tested against other codes or models
qx, qy = 0.2 * cos(pi/6.), 0.2 * sin(pi/6.)
tests = [[{}, 0.2, 0.533149288477],
         [{}, [0.2], [0.533149288477]],
         [{'theta':10.0, 'phi':20.0}, (qx, qy), 0.0853299803222],
         [{'theta':10.0, 'phi':20.0}, [(qx, qy)], [0.0853299803222]],
        ]
del qx, qy  # not necessary to delete, but cleaner



r"""
Definition
----------

The scattering intensity $I(q)$ is calculated as

.. math::

    I(q) = \begin{cases}
    A q^{-m1} + \text{background} & q <= q_c \\
    C q^{-m2} + \text{background} & q > q_c
    \end{cases}

where $q_c$ = the location of the crossover from one slope to the other,
$A$ = the scaling coefficent that sets the overall intensity of the lower Q
power law region, $m1$ = power law exponent at low Q, and $m2$ = power law
exponent at high Q.  The scaling of the second power law region (coefficent C)
is then automatically scaled to match the first by following formula:

.. math::
    C = \frac{A q_c^{m2}}{q_c^{m1}}

.. note::
    Be sure to enter the power law exponents as positive values!

For 2D data the scattering intensity is calculated in the same way as 1D,
where the $q$ vector is defined as

.. math::

    q = \sqrt{q_x^2 + q_y^2}


References
----------

None.

**Author:** NIST IGOR/DANSE **on:** pre 2010

**Last Modified by:** Wojciech Wpotrzebowski **on:** February 18, 2016

**Last Reviewed by:** Paul Butler **on:** March 21, 2016
"""

from numpy import inf, power, empty, errstate

name = "two_power_law"
title = "This model calculates an empirical functional form for SAS data \
characterized by two power laws."
description = """
            I(q) = coef_A*pow(qval,-1.0*power1) + background for q<=q_c
            =C*pow(qval,-1.0*power2) + background for q>q_c
            where C=coef_A*pow(q_c,-1.0*power1)/pow(q_c,-1.0*power2).

            coef_A = scaling coefficent
            q_c = crossover location [1/A]
            power_1 (=m1) = power law exponent at low Q
            power_2 (=m2) = power law exponent at high Q
            background = Incoherent background [1/cm]
        """
category = "shape-independent"

# pylint: disable=bad-whitespace, line-too-long
#   ["name", "units", default, [lower, upper], "type", "description"],
parameters = [
    ["coefficent_1", "",       1.0, [-inf, inf], "", "coefficent A in low Q region"],
    ["crossover",    "1/Ang",  0.04,[0, inf],    "", "crossover location"],
    ["power_1",      "",       1.0, [0, inf],    "", "power law exponent at low Q"],
    ["power_2",      "",       4.0, [0, inf],    "", "power law exponent at high Q"],
    ]
# pylint: enable=bad-whitespace, line-too-long


def Iq(q,
       coefficent_1=1.0,
       crossover=0.04,
       power_1=1.0,
       power_2=4.0,
      ):

    """
    :param q:                   Input q-value (float or [float, float])
    :param coefficent_1:        Scaling coefficent at low Q
    :param crossover:           Crossover location
    :param power_1:             Exponent of power law function at low Q
    :param power_2:             Exponent of power law function at high Q
    :return:                    Calculated intensity
    """
    result= empty(q.shape, 'd')
    index = (q <= crossover)
    with errstate(divide='ignore'):
        coefficent_2 = coefficent_1 * power(crossover, power_2 - power_1)
        result[index] = coefficent_1 * power(q[index], -power_1)
        result[~index] = coefficent_2 * power(q[~index], -power_2)
    return result

Iq.vectorized = True  # Iq accepts an array of q values

def random():
    import numpy as np
    coefficient_1 = 1
    crossover = 10**np.random.uniform(-3, -1)
    power_1 = np.random.uniform(1, 6)
    power_2 = np.random.uniform(1, 6)
    pars = dict(
        scale=1, #background=0,
        coefficient_1=coefficient_1,
        crossover=crossover,
        power_1=power_1,
        power_2=power_2,
    )
    return pars

demo = dict(scale=1, background=0.0,
            coefficent_1=1.0,
            crossover=0.04,
            power_1=1.0,
            power_2=4.0)

tests = [
    # Accuracy tests based on content in test/utest_extra_models.py
    [{'coefficent_1':     1.0,
      'crossover':  0.04,
      'power_1':    1.0,
      'power_2':    4.0,
      'background': 0.0,
     }, 0.001, 1000],

    [{'coefficent_1':     1.0,
      'crossover':  0.04,
      'power_1':    1.0,
      'power_2':    4.0,
      'background': 0.0,
     }, 0.150141, 0.125945],

    [{'coefficent_1':    1.0,
      'crossover':  0.04,
      'power_1':    1.0,
      'power_2':    4.0,
      'background': 0.0,
     }, 0.442528, 0.00166884],

    [{'coefficent_1':    1.0,
      'crossover':  0.04,
      'power_1':    1.0,
      'power_2':    4.0,
      'background': 0.0,
     }, (0.442528, 0.00166884), 0.00166884],

]
